# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/default_bank_definitions.py
# Compiled at: 2023-12-21 15:35:35
# Size of source mod 2**32: 132452 bytes
from ableton.v2.base.collection import IndexedDict

from . import BANK_MAIN_KEY, BANK_PARAMETERS_KEY, use

BANK_DEFINITIONS = {}
RACK_BANKS = IndexedDict(
    (
        (
            "Macros",
            {
                BANK_PARAMETERS_KEY: (
                    "Macro 1",
                    "Macro 2",
                    "Macro 3",
                    "Macro 4",
                    "Macro 5",
                    "Macro 6",
                    "Macro 7",
                    "Macro 8",
                ),
            },
        ),
        (
            "Macros 2",
            {
                BANK_PARAMETERS_KEY: (
                    "Macro 9",
                    "Macro 10",
                    "Macro 11",
                    "Macro 12",
                    "Macro 13",
                    "Macro 14",
                    "Macro 15",
                    "Macro 16",
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["AudioEffectGroupDevice"] = RACK_BANKS
BANK_DEFINITIONS["MidiEffectGroupDevice"] = RACK_BANKS
BANK_DEFINITIONS["InstrumentGroupDevice"] = RACK_BANKS
BANK_DEFINITIONS["DrumGroupDevice"] = RACK_BANKS
BANK_DEFINITIONS["Amp"] = IndexedDict(
    (
        (
            "Global",
            {BANK_PARAMETERS_KEY: ("Amp Type", "Bass", "Middle", "Treble", "Presence", "Gain", "Volume", "Dry/Wet")},
        ),
        ("Dual/Mono", {BANK_PARAMETERS_KEY: ("Dual Mono", "", "", "", "", "", "", "")}),
    ),
)
BANK_DEFINITIONS["UltraAnalog"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    "F1 Freq",
                    "F1 Resonance",
                    "OSC1 Shape",
                    "OSC1 Octave",
                    "OSC2 Shape",
                    "OSC2 Octave",
                    "OSC2 Detune",
                    "Volume",
                ),
            },
        ),
        (
            "Oscillators",
            {
                BANK_PARAMETERS_KEY: (
                    "OSC1 Level",
                    "OSC1 Octave",
                    "OSC1 Semi",
                    "OSC1 Shape",
                    "OSC2 Level",
                    "OSC2 Octave",
                    "OSC2 Semi",
                    "OSC2 Shape",
                ),
            },
        ),
        (
            "Filters",
            {
                BANK_PARAMETERS_KEY: (
                    "OSC1 Balance",
                    "F1 Freq",
                    "F1 Resonance",
                    "F1 Type",
                    "OSC2 Balance",
                    "F2 Freq",
                    "F2 Resonance",
                    "F2 Type",
                ),
            },
        ),
        (
            "Filter Envelopes",
            {
                BANK_PARAMETERS_KEY: (
                    "FEG1 Attack",
                    "FEG1 Decay",
                    "FEG1 Sustain",
                    "FEG1 Rel",
                    "FEG2 Attack",
                    "FEG2 Decay",
                    "FEG2 Sustain",
                    "FEG2 Rel",
                ),
            },
        ),
        (
            "Filter Modulation",
            {
                BANK_PARAMETERS_KEY: (
                    "F1 On/Off",
                    "F1 Freq < LFO",
                    "F1 Freq < Env",
                    "F1 Res < LFO",
                    "F2 On/Off",
                    "F2 Freq < LFO",
                    "F2 Freq < Env",
                    "F2 Res < LFO",
                ),
            },
        ),
        (
            "Volume Envelopes",
            {
                BANK_PARAMETERS_KEY: (
                    "AEG1 Attack",
                    "AEG1 Decay",
                    "AEG1 Sustain",
                    "AEG1 Rel",
                    "AEG2 Attack",
                    "AEG2 Decay",
                    "AEG2 Sustain",
                    "AEG2 Rel",
                ),
            },
        ),
        (
            "Mix",
            {
                BANK_PARAMETERS_KEY: (
                    "AMP1 Level",
                    "AMP1 Pan",
                    "LFO1 Shape",
                    use("LFO1 Speed").if_parameter("LFO1 Sync").has_value("Hertz").else_use("LFO1 SncRate"),
                    "AMP2 Level",
                    "AMP2 Pan",
                    "LFO2 Shape",
                    use("LFO2 Speed").if_parameter("LFO2 Sync").has_value("Hertz").else_use("LFO2 SncRate"),
                ),
            },
        ),
        (
            "Output",
            {
                BANK_PARAMETERS_KEY: (
                    "Volume",
                    "Noise On/Off",
                    "Noise Level",
                    "Noise Color",
                    "Unison On/Off",
                    "Unison Detune",
                    "Vib On/Off",
                    "Vib Amount",
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["MidiArpeggiator"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    "Style",
                    use("Synced Rate").if_parameter("Sync On").has_value("On").else_use("Free Rate"),
                    "Gate",
                    "Offset",
                    "Hold On",
                    "Tranpose Key",
                    "Transp. Steps",
                    "Transp. Dist.",
                ),
            },
        ),
        (
            "Rhythm",
            {
                BANK_PARAMETERS_KEY: (
                    "Sync On",
                    use("Synced Rate").if_parameter("Sync On").has_value("On").else_use("Free Rate"),
                    "Groove",
                    "Offset",
                    "Repeats",
                    "Gate",
                    "Retrigger Mode",
                    "Ret. Interval",
                ),
            },
        ),
        (
            "Pitch/Vel.",
            {
                BANK_PARAMETERS_KEY: (
                    "Tranpose Mode",
                    "Tranpose Key",
                    "Transp. Steps",
                    "Transp. Dist.",
                    "Velocity On",
                    "Vel. Retrigger",
                    "Velocity Decay",
                    "Velocity Target",
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["AutoFilter"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    use("Filter Type").if_parameter("Filter Type").is_available(True).else_use("Filter Type (Legacy)"),
                    use("Frequency"),
                    use("Resonance").if_parameter("Resonance").is_available(True).else_use("Resonance (Legacy)"),
                    use("Filter Circuit - LP/HP")
                    .if_parameter("Filter Type")
                    .has_value_in(("Lowpass", "Highpass"))
                    .else_use("Filter Circuit - BP/NO/Morph"),
                    use("Morph")
                    .if_parameter("Filter Type")
                    .has_value("Morph")
                    .else_use("Drive")
                    .if_parameter("Drive")
                    .is_available(True),
                    "LFO Amount",
                    "LFO Sync",
                    use("LFO Frequency").if_parameter("LFO Sync").has_value("Free").else_use("LFO Sync Rate"),
                ),
            },
        ),
        (
            "Envelope",
            {
                BANK_PARAMETERS_KEY: (
                    use("Filter Type").if_parameter("Filter Type").is_available(True).else_use("Filter Type (Legacy)"),
                    use("Frequency"),
                    use("Resonance").if_parameter("Resonance").is_available(True).else_use("Resonance (Legacy)"),
                    use("Morph")
                    .if_parameter("Filter Type")
                    .has_value("Morph")
                    .else_use("Drive")
                    .if_parameter("Drive")
                    .is_available(True),
                    use("Slope").if_parameter("Slope").is_available(True),
                    "Env. Attack",
                    "Env. Release",
                    "Env. Modulation",
                ),
            },
        ),
        (
            "LFO",
            {
                BANK_PARAMETERS_KEY: (
                    "LFO Amount",
                    "LFO Waveform",
                    "LFO Sync",
                    use("LFO Frequency").if_parameter("LFO Sync").has_value("Free").else_use("LFO Sync Rate"),
                    use("LFO Offset").if_parameter("LFO Sync").has_value("Sync").else_use("LFO Stereo Mode"),
                    use("LFO Phase")
                    .if_parameter("LFO Sync")
                    .has_value("Sync")
                    .or_parameter("LFO Stereo Mode")
                    .has_value("Phase")
                    .else_use("LFO Spin"),
                    "LFO Quantize On",
                    "LFO Quantize Rate",
                ),
            },
        ),
        ("Sidechain", {BANK_PARAMETERS_KEY: ("S/C On", "S/C Mix", "S/C Gain", "", "", "", "", "")}),
    ),
)
BANK_DEFINITIONS["AutoPan"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    "Amount",
                    "Shape",
                    "Invert",
                    "Waveform",
                    "LFO Type",
                    use("Sync Rate").if_parameter("LFO Type").has_value("Beats").else_use("Frequency"),
                    use("Stereo Mode").if_parameter("LFO Type").has_value("Frequency").else_use("Offset"),
                    use("Width (Random)")
                    .if_parameter("Waveform")
                    .has_value("S&H Width")
                    .else_use("Phase")
                    .if_parameter("LFO Type")
                    .has_value("Beats")
                    .else_use("Spin")
                    .if_parameter("Stereo Mode")
                    .has_value("Spin")
                    .else_use("Phase"),
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["BeatRepeat"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    "Grid",
                    "Interval",
                    "Offset",
                    "Gate",
                    "Pitch",
                    "Pitch Decay",
                    "Variation",
                    "Chance",
                ),
            },
        ),
        (
            "Filt/Mix",
            {
                BANK_PARAMETERS_KEY: (
                    "Filter On",
                    "Filter Freq",
                    "Filter Width",
                    "",
                    "Mix Type",
                    "Volume",
                    "Decay",
                    "Chance",
                ),
            },
        ),
        (
            "Repeat Rate",
            {
                BANK_PARAMETERS_KEY: (
                    "Repeat",
                    "Interval",
                    "Offset",
                    "Gate",
                    "Grid",
                    "Block Triplets",
                    "Variation",
                    "Variation Type",
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["Cabinet"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    "Cabinet Type",
                    "Microphone Type",
                    "Microphone Position",
                    "Dual Mono",
                    "",
                    "",
                    "",
                    "Dry/Wet",
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["ChannelEq"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    "Highpass On",
                    use("Low Gain").with_name("Low"),
                    use("Mid Gain").with_name("Mid"),
                    "Mid Freq",
                    use("High Gain").with_name("High"),
                    use("Gain").with_name("Output"),
                    "",
                    "",
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["MidiChord"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    "Shift1",
                    "Shift2",
                    "Shift3",
                    "Shift4",
                    "Shift5",
                    "Velocity5",
                    "Shift6",
                    "Velocity6",
                ),
            },
        ),
        ("Shift", {BANK_PARAMETERS_KEY: ("Shift1", "Shift2", "Shift3", "Shift4", "Shift5", "Shift6", "", "")}),
        (
            "Shift %",
            {
                BANK_PARAMETERS_KEY: (
                    "Velocity1",
                    "Velocity2",
                    "Velocity3",
                    "Velocity4",
                    "Velocity5",
                    "Velocity6",
                    "",
                    "",
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["Chorus"] = IndexedDict(
    (
        (
            "Chorus",
            {
                BANK_PARAMETERS_KEY: (
                    "LFO Amount",
                    "LFO Rate",
                    "Delay 1 Time",
                    "Delay 1 HiPass",
                    "Delay 2 Mode",
                    "Delay 2 Time",
                    "Feedback",
                    "Dry/Wet",
                ),
            },
        ),
        ("Other", {BANK_PARAMETERS_KEY: ("LFO Extend On", "Polarity", "Link On", "", "", "", "", "")}),
    ),
)
BANK_DEFINITIONS["Chorus2"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    "Mode",
                    "Rate",
                    "Amount",
                    "Feedback",
                    use("Offset").if_parameter("Mode").has_value("Vibrato").else_use("Width"),
                    "Gain",
                    "Warmth",
                    "Dry/Wet",
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["Collision"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    "Res 1 Brightness",
                    "Res 1 Type",
                    "Mallet Stiffness",
                    "Mallet Noise Amount",
                    use("Res 1 Opening")
                    .if_parameter("Res 1 Type")
                    .has_value_in(("Tube", "Pipe"))
                    .else_use("Res 1 Inharmonics"),
                    "Res 1 Decay",
                    "Res 1 Tune",
                    "Volume",
                ),
            },
        ),
        (
            "Mallet",
            {
                BANK_PARAMETERS_KEY: (
                    "Mallet On/Off",
                    "Mallet Volume",
                    "Mallet Noise Amount",
                    "Mallet Stiffness",
                    "Mallet Noise Color",
                    "",
                    "",
                    "",
                ),
            },
        ),
        (
            "Noise",
            {
                BANK_PARAMETERS_KEY: (
                    "Noise Volume",
                    "Noise Filter Type",
                    "Noise Filter Freq",
                    "Noise Filter Q",
                    "Noise Attack",
                    "Noise Decay",
                    "Noise Sustain",
                    "Noise Release",
                ),
            },
        ),
        (
            "Resonator 1, Set A",
            {
                BANK_PARAMETERS_KEY: (
                    "Res 1 Decay",
                    use("Res 1 Radius")
                    .if_parameter("Res 1 Type")
                    .has_value_in(("Tube", "Pipe"))
                    .else_use("Res 1 Material"),
                    "Res 1 Type",
                    "Res 1 Quality",
                    "Res 1 Tune",
                    "Res 1 Fine Tune",
                    "Res 1 Pitch Env.",
                    "Res 1 Pitch Env. Time",
                ),
            },
        ),
        (
            "Resonator 1, Set B",
            {
                BANK_PARAMETERS_KEY: (
                    "Res 1 Listening L",
                    "Res 1 Listening R",
                    "Res 1 Hit",
                    "Res 1 Brightness",
                    "Res 1 Decay",
                    use("Res 1 Radius")
                    .if_parameter("Res 1 Type")
                    .has_value_in(("Tube", "Pipe"))
                    .else_use("Res 1 Material"),
                    use("Res 1 Opening")
                    .if_parameter("Res 1 Type")
                    .has_value_in(("Tube", "Pipe"))
                    .else_use("Res 1 Inharmonics"),
                    "Res 1 Ratio",
                ),
            },
        ),
        (
            "Resonator 2, Set A",
            {
                BANK_PARAMETERS_KEY: (
                    "Res 2 Decay",
                    use("Res 2 Radius")
                    .if_parameter("Res 2 Type")
                    .has_value_in(("Tube", "Pipe"))
                    .else_use("Res 2 Material"),
                    "Res 2 Type",
                    "Res 2 Quality",
                    "Res 2 Tune",
                    "Res 2 Fine Tune",
                    "Res 2 Pitch Env.",
                    "Res 2 Pitch Env. Time",
                ),
            },
        ),
        (
            "Resonator 2, Set B",
            {
                BANK_PARAMETERS_KEY: (
                    "Res 2 Listening L",
                    "Res 2 Listening R",
                    "Res 2 Hit",
                    "Res 2 Brightness",
                    "Res 2 Decay",
                    use("Res 2 Radius")
                    .if_parameter("Res 2 Type")
                    .has_value_in(("Tube", "Pipe"))
                    .else_use("Res 2 Material"),
                    use("Res 2 Opening")
                    .if_parameter("Res 2 Type")
                    .has_value_in(("Tube", "Pipe"))
                    .else_use("Res 2 Inharmonics"),
                    "Res 2 Ratio",
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["Compressor2"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    "Threshold",
                    use("Expansion Ratio").if_parameter("Model").has_value("Expand").else_use("Ratio"),
                    "Model",
                    "Knee",
                    "Attack",
                    "Release",
                    "Dry/Wet",
                    "Output Gain",
                ),
            },
        ),
        (
            "Sidechain",
            {
                BANK_PARAMETERS_KEY: (
                    "S/C On",
                    "S/C Gain",
                    "S/C Mix",
                    "S/C Listen",
                    "S/C EQ On",
                    "S/C EQ Type",
                    "S/C EQ Freq",
                    use("S/C EQ Gain")
                    .if_parameter("S/C EQ Type")
                    .has_value_in(("Low Shelf", "High Shelf", "Bell"))
                    .else_use("S/C EQ Q"),
                ),
            },
        ),
        (
            "Global",
            {
                BANK_PARAMETERS_KEY: (
                    "Auto Release On/Off",
                    "Env Mode",
                    use("Makeup").if_parameter("S/C On").has_value("Off"),
                    "LookAhead",
                    "",
                    "",
                    "Dry/Wet",
                    "Output Gain",
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["Corpus"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    "Resonance Type",
                    "Brightness",
                    "Decay",
                    use("Radius").if_parameter("Resonance Type").has_value_in(("Tube", "Pipe")).else_use("Material"),
                    use("Opening")
                    .if_parameter("Resonance Type")
                    .has_value_in(("Tube", "Pipe"))
                    .else_use("Inharmonics"),
                    "Ratio",
                    use("Transpose").if_parameter("MIDI Frequency").has_value("On").else_use("Tune"),
                    use("Dry Wet").with_name("Dry/Wet"),
                ),
            },
        ),
        (
            "Body",
            {
                BANK_PARAMETERS_KEY: (
                    "Resonance Type",
                    "Ratio",
                    "Decay",
                    use("Radius").if_parameter("Resonance Type").has_value_in(("Tube", "Pipe")).else_use("Material"),
                    use("Opening")
                    .if_parameter("Resonance Type")
                    .has_value_in(("Tube", "Pipe"))
                    .else_use("Inharmonics"),
                    "Listening L",
                    "Listening R",
                    "Hit",
                ),
            },
        ),
        (
            "LFO",
            {
                BANK_PARAMETERS_KEY: (
                    "LFO On/Off",
                    "LFO Shape",
                    "LFO Amount",
                    "LFO Sync",
                    use("LFO Rate").if_parameter("LFO Sync").has_value("Free").else_use("LFO Sync Rate"),
                    use("LFO Stereo Mode").if_parameter("LFO Sync").has_value("Free").else_use("Offset"),
                    use("Phase")
                    .if_parameter("LFO Sync")
                    .has_value("Sync")
                    .or_parameter("LFO Stereo Mode")
                    .has_value("Phase")
                    .else_use("Spin"),
                    "",
                ),
            },
        ),
        (
            "Tune & Sidechain",
            {
                BANK_PARAMETERS_KEY: (
                    "MIDI Frequency",
                    "MIDI Mode",
                    use("Transpose").if_parameter("MIDI Frequency").has_value("On").else_use("Tune"),
                    "Fine",
                    "Spread",
                    "Brightness",
                    "Note Off",
                    "Off Decay",
                ),
            },
        ),
        (
            "Filter & Mix",
            {
                BANK_PARAMETERS_KEY: (
                    "Filter On/Off",
                    "Mid Freq",
                    "Width",
                    "Resonator Quality",
                    "Bleed",
                    "Gain",
                    use("Dry Wet").with_name("Dry/Wet"),
                    "",
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["Delay"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    use("L Sync Enum").with_name("DlayMd L"),
                    use("L 16th").if_parameter("L Sync").has_value("On").else_use("L Time"),
                    "L Offset",
                    use("R Sync Enum")
                    .with_name("DlayMd R")
                    .if_parameter("Link")
                    .has_value("Off")
                    .else_use("L Sync Enum")
                    .with_name("DlayMd L"),
                    use("R 16th")
                    .if_parameter("R Sync")
                    .has_value("On")
                    .and_parameter("Link")
                    .has_value("Off")
                    .else_use("R Time")
                    .if_parameter("Link")
                    .has_value("Off")
                    .else_use("L 16th")
                    .if_parameter("L Sync")
                    .has_value("On")
                    .else_use("L Time"),
                    "R Offset",
                    "Feedback",
                    "Dry/Wet",
                ),
            },
        ),
        (
            "Time",
            {
                BANK_PARAMETERS_KEY: (
                    use("Delay Mode").with_name("DlayMd"),
                    "Link",
                    "Ping Pong",
                    "L Offset",
                    "R Offset",
                    "Freeze",
                    "Feedback",
                    "Dry/Wet",
                ),
            },
        ),
        (
            "Filter/Mod",
            {
                BANK_PARAMETERS_KEY: (
                    "Filter On",
                    "Filter Freq",
                    "Filter Width",
                    use("Mod Freq").with_name("Filter Mod Rate"),
                    "Filter < Mod",
                    "Dly < Mod",
                    "Feedback",
                    "Dry/Wet",
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["Drift"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    "Osc 1 Wave",
                    "Osc 1 Shape",
                    "Osc 1 Oct",
                    "Osc 1 Gain",
                    "Osc 2 Gain",
                    "LP Freq",
                    "LP Reso",
                    "Volume",
                ),
            },
        ),
        (
            "Oscillator",
            {
                BANK_PARAMETERS_KEY: (
                    "Osc 1 Wave",
                    "Osc 1 Shape",
                    "Osc 1 Oct",
                    "Osc 1 Gain",
                    "Osc 2 Wave",
                    "Osc 2 Detune",
                    "Osc 2 Oct",
                    "Osc 2 Gain",
                ),
            },
        ),
        (
            "Filter",
            {
                BANK_PARAMETERS_KEY: (
                    "LP Type",
                    "LP Freq",
                    "LP Reso",
                    "HP Freq",
                    "LP Mod Src 1",
                    "LP Mod Amt 1",
                    "LP Mod Src 2",
                    "LP Mod Amt 2",
                ),
            },
        ),
        (
            "Envelopes",
            {
                BANK_PARAMETERS_KEY: (
                    "Env 1 Attack",
                    "Env 1 Decay",
                    "Env 1 Sustain",
                    "Env 1 Release",
                    "Env 2 Attack",
                    "Env 2 Decay",
                    "Env 2 Sustain",
                    "Env 2 Release",
                ),
            },
        ),
        (
            "LFO",
            {
                BANK_PARAMETERS_KEY: (
                    "LFO Wave",
                    "LFO Time Mode",
                    use("LFO Rate")
                    .if_parameter("LFO Time Mode")
                    .has_value("Freq")
                    .else_use("LFO Ratio")
                    .if_parameter("LFO Time Mode")
                    .has_value("Ratio")
                    .else_use("LFO Time")
                    .if_parameter("LFO Time Mode")
                    .has_value("Time")
                    .else_use("LFO Synced"),
                    "LFO Amt",
                    "Cyc Env Tilt",
                    "Cyc Env Hold",
                    "Cyc Env Time Mode",
                    use("Cyc Env Rate")
                    .if_parameter("Cyc Env Time Mode")
                    .has_value("Freq")
                    .else_use("Cyc Env Ratio")
                    .if_parameter("Cyc Env Time Mode")
                    .has_value("Ratio")
                    .else_use("Cyc Env Time")
                    .if_parameter("Cyc Env Time Mode")
                    .has_value("Time")
                    .else_use("Cyc Env Synced"),
                ),
            },
        ),
        (
            "Modulation",
            {
                BANK_PARAMETERS_KEY: (
                    "Osc 1 Shape Mod Amt",
                    "Pitch Mod Amt 1",
                    "Pitch Mod Amt 2",
                    "LP Mod Amt 1",
                    "LP Mod Amt 2",
                    "LFO Mod Amt",
                    "Vel > Vol",
                    "",
                ),
            },
        ),
        (
            "Global",
            {
                BANK_PARAMETERS_KEY: (
                    "Voice Mode",
                    use("")
                    .if_parameter("Voice Mode")
                    .has_value("Poly")
                    .else_use("Thickness")
                    .if_parameter("Voice Mode")
                    .has_value("Mono")
                    .else_use("Spread")
                    .if_parameter("Voice Mode")
                    .has_value("Stereo")
                    .else_use("Strength")
                    .if_parameter("Voice Mode")
                    .has_value("Unison"),
                    "Voice Count",
                    "Glide Time",
                    "Drift",
                    "Transpose",
                    "Noise Gain",
                    "Volume",
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["DrumBuss"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    "Drive",
                    "Drive Type",
                    "Crunch",
                    "Boom Amt",
                    "Trim",
                    "Damping Freq",
                    "Output Gain",
                    "Dry/Wet",
                ),
            },
        ),
        (
            "Drive",
            {
                BANK_PARAMETERS_KEY: (
                    "Drive",
                    "Drive Type",
                    "Transients",
                    "Crunch",
                    "Boom Freq",
                    "Boom Amt",
                    "Boom Decay",
                    "Boom Audition",
                ),
            },
        ),
        (
            "Gain",
            {BANK_PARAMETERS_KEY: ("Trim", "Output Gain", "Dry/Wet", "Compressor On", "Damping Freq", "", "", "")},
        ),
    ),
)
BANK_DEFINITIONS["Tube"] = IndexedDict(
    (
        (
            "Character",
            {BANK_PARAMETERS_KEY: ("Drive", "Tube Type", "Bias", "Tone", "Attack", "Release", "Envelope", "Dry/Wet")},
        ),
        ("Output", {BANK_PARAMETERS_KEY: ("", "", "", "", "", "", "Output", "Dry/Wet")}),
    ),
)
BANK_DEFINITIONS["FilterEQ3"] = IndexedDict(
    (
        ("EQ", {BANK_PARAMETERS_KEY: ("LowOn", "MidOn", "HighOn", "GainLo", "GainMid", "GainHi", "FreqLo", "FreqHi")}),
        ("Slope", {BANK_PARAMETERS_KEY: ("Slope", "", "", "", "", "", "", "")}),
    ),
)
BANK_DEFINITIONS["Eq8"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    "1 Frequency A",
                    "1 Gain A",
                    "2 Frequency A",
                    "2 Gain A",
                    "3 Frequency A",
                    "3 Gain A",
                    "4 Frequency A",
                    "4 Gain A",
                ),
            },
        ),
        (
            "Band On/Off",
            {
                BANK_PARAMETERS_KEY: (
                    "1 Filter On A",
                    "2 Filter On A",
                    "3 Filter On A",
                    "4 Filter On A",
                    "5 Filter On A",
                    "6 Filter On A",
                    "7 Filter On A",
                    "8 Filter On A",
                ),
            },
        ),
        (
            "Frequency",
            {
                BANK_PARAMETERS_KEY: (
                    "1 Frequency A",
                    "2 Frequency A",
                    "3 Frequency A",
                    "4 Frequency A",
                    "5 Frequency A",
                    "6 Frequency A",
                    "7 Frequency A",
                    "8 Frequency A",
                ),
            },
        ),
        (
            "Gain",
            {
                BANK_PARAMETERS_KEY: (
                    "1 Gain A",
                    "2 Gain A",
                    "3 Gain A",
                    "4 Gain A",
                    "5 Gain A",
                    "6 Gain A",
                    "7 Gain A",
                    "8 Gain A",
                ),
            },
        ),
        (
            "Resonance",
            {
                BANK_PARAMETERS_KEY: (
                    "1 Resonance A",
                    "2 Resonance A",
                    "3 Resonance A",
                    "4 Resonance A",
                    "5 Resonance A",
                    "6 Resonance A",
                    "7 Resonance A",
                    "8 Resonance A",
                ),
            },
        ),
        (
            "Filter Type",
            {
                BANK_PARAMETERS_KEY: (
                    "1 Filter Type A",
                    "2 Filter Type A",
                    "3 Filter Type A",
                    "4 Filter Type A",
                    "5 Filter Type A",
                    "6 Filter Type A",
                    "7 Filter Type A",
                    "8 Filter Type A",
                ),
            },
        ),
        ("Output", {BANK_PARAMETERS_KEY: ("Adaptive Q", "", "", "", "", "", "Scale", "Output Gain")}),
        (
            "EQs 3-5",
            {
                BANK_PARAMETERS_KEY: (
                    "3 Gain A",
                    "3 Frequency A",
                    "3 Resonance A",
                    "4 Gain A",
                    "4 Frequency A",
                    "4 Resonance A",
                    "5 Gain A",
                    "5 Frequency A",
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["Echo"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    use("L Time")
                    .with_name("M Time")
                    .if_parameter("Channel Mode")
                    .has_value("Mid/Side")
                    .and_parameter("L Sync")
                    .has_value("Off")
                    .else_use("L Time")
                    .if_parameter("L Sync")
                    .has_value("Off")
                    .else_use("L 16th")
                    .with_name("M 16th")
                    .if_parameter("Channel Mode")
                    .has_value("Mid/Side")
                    .and_parameter("L Sync Mode")
                    .has_value("16th")
                    .else_use("L 16th")
                    .if_parameter("L Sync Mode")
                    .has_value("16th")
                    .else_use("L Division")
                    .with_name("M Division")
                    .if_parameter("Channel Mode")
                    .has_value("Mid/Side")
                    .else_use("L Division"),
                    use("L Sync")
                    .with_name("M Sync")
                    .if_parameter("Channel Mode")
                    .has_value("Mid/Side")
                    .else_use("L Sync"),
                    use("L Sync Mode")
                    .with_name("M Sync Mode")
                    .if_parameter("Channel Mode")
                    .has_value("Mid/Side")
                    .else_use("L Sync Mode"),
                    "HP Freq",
                    "LP Freq",
                    "Feedback",
                    "Input Gain",
                    use("Dry Wet").with_name("Dry/Wet"),
                ),
            },
        ),
        (
            "L/Mid",
            {
                BANK_PARAMETERS_KEY: (
                    use("L Time")
                    .with_name("M Time")
                    .if_parameter("Channel Mode")
                    .has_value("Mid/Side")
                    .and_parameter("L Sync")
                    .has_value("Off")
                    .else_use("L Time")
                    .if_parameter("L Sync")
                    .has_value("Off")
                    .else_use("L 16th")
                    .with_name("M 16th")
                    .if_parameter("Channel Mode")
                    .has_value("Mid/Side")
                    .and_parameter("L Sync Mode")
                    .has_value("16th")
                    .else_use("L 16th")
                    .if_parameter("L Sync Mode")
                    .has_value("16th")
                    .else_use("L Division")
                    .with_name("M Division")
                    .if_parameter("Channel Mode")
                    .has_value("Mid/Side")
                    .else_use("L Division"),
                    use("L Sync")
                    .with_name("M Sync")
                    .if_parameter("Channel Mode")
                    .has_value("Mid/Side")
                    .else_use("L Sync"),
                    use("L Sync Mode")
                    .with_name("M Sync Mode")
                    .if_parameter("Channel Mode")
                    .has_value("Mid/Side")
                    .else_use("L Sync Mode"),
                    use("L Offset")
                    .with_name("M Offset")
                    .if_parameter("Channel Mode")
                    .has_value("Mid/Side")
                    .else_use("L Offset"),
                    "Link",
                    "Feedback",
                    "Feedback Inv",
                    use("Dry Wet").with_name("Dry/Wet"),
                ),
            },
        ),
        (
            "R/Side",
            {
                BANK_PARAMETERS_KEY: (
                    use("R Time")
                    .with_name("S Time")
                    .if_parameter("Channel Mode")
                    .has_value("Mid/Side")
                    .and_parameter("R Sync")
                    .has_value("Off")
                    .and_parameter("Link")
                    .has_value("Off")
                    .else_use("R Time")
                    .if_parameter("R Sync")
                    .has_value("Off")
                    .and_parameter("Link")
                    .has_value("Off")
                    .else_use("R 16th")
                    .with_name("S 16th")
                    .if_parameter("Channel Mode")
                    .has_value("Mid/Side")
                    .and_parameter("R Sync Mode")
                    .has_value("16th")
                    .and_parameter("Link")
                    .has_value("Off")
                    .else_use("R 16th")
                    .if_parameter("R Sync Mode")
                    .has_value("16th")
                    .and_parameter("Link")
                    .has_value("Off")
                    .else_use("R Division")
                    .with_name("S Division")
                    .if_parameter("Channel Mode")
                    .has_value("Mid/Side")
                    .and_parameter("Link")
                    .has_value("Off")
                    .else_use("R Division")
                    .if_parameter("Link")
                    .has_value("Off")
                    .else_use("L Time")
                    .with_name("M Time")
                    .if_parameter("Channel Mode")
                    .has_value("Mid/Side")
                    .and_parameter("L Sync")
                    .has_value("Off")
                    .else_use("L Time")
                    .if_parameter("L Sync")
                    .has_value("Off")
                    .else_use("L 16th")
                    .with_name("M 16th")
                    .if_parameter("Channel Mode")
                    .has_value("Mid/Side")
                    .and_parameter("L Sync Mode")
                    .has_value("16th")
                    .else_use("L 16th")
                    .if_parameter("L Sync Mode")
                    .has_value("16th")
                    .else_use("L Division")
                    .with_name("M Division")
                    .if_parameter("Channel Mode")
                    .has_value("Mid/Side")
                    .else_use("L Division"),
                    use("R Sync")
                    .with_name("S Sync")
                    .if_parameter("Channel Mode")
                    .has_value("Mid/Side")
                    .and_parameter("Link")
                    .has_value("Off")
                    .else_use("R Sync")
                    .if_parameter("Link")
                    .has_value("Off")
                    .else_use("L Sync")
                    .with_name("M Sync")
                    .if_parameter("Channel Mode")
                    .has_value("Mid/Side")
                    .else_use("L Sync"),
                    use("R Sync Mode")
                    .with_name("S Sync Mode")
                    .if_parameter("Channel Mode")
                    .has_value("Mid/Side")
                    .and_parameter("Link")
                    .has_value("Off")
                    .else_use("R Sync Mode")
                    .if_parameter("Link")
                    .has_value("Off")
                    .else_use("L Sync Mode")
                    .with_name("M Sync Mode")
                    .if_parameter("Channel Mode")
                    .has_value("Mid/Side")
                    .else_use("L Sync Mode"),
                    use("R Offset")
                    .with_name("S Offset")
                    .if_parameter("Channel Mode")
                    .has_value("Mid/Side")
                    .else_use("R Offset"),
                    "Link",
                    "Feedback",
                    "Feedback Inv",
                    use("Dry Wet").with_name("Dry/Wet"),
                ),
            },
        ),
        (
            "Global",
            {
                BANK_PARAMETERS_KEY: (
                    "Repitch",
                    "Channel Mode",
                    "Stereo Width",
                    "",
                    "Clip Dry",
                    "Input Gain",
                    "Output Gain",
                    use("Dry Wet").with_name("Dry/Wet"),
                ),
            },
        ),
        (
            "Filter",
            {
                BANK_PARAMETERS_KEY: (
                    "Filter On",
                    "HP Freq",
                    "HP Res",
                    "LP Freq",
                    "LP Res",
                    "Input Gain",
                    "Output Gain",
                    use("Dry Wet").with_name("Dry/Wet"),
                ),
            },
        ),
        (
            "Modulation",
            {
                BANK_PARAMETERS_KEY: (
                    "Mod Wave",
                    "Mod Sync",
                    use("Mod Rate").if_parameter("Mod Sync").has_value("On").else_use("Mod Freq"),
                    "Mod Phase",
                    "Env Mix",
                    "Dly < Mod",
                    "Flt < Mod",
                    "Mod 4x",
                ),
            },
        ),
        ("Reverb", {BANK_PARAMETERS_KEY: ("Reverb Level", "Reverb Loc", "Reverb Decay", "", "", "", "", "")}),
        (
            "Gate/Ducking",
            {
                BANK_PARAMETERS_KEY: (
                    "Gate On",
                    "Gate Thr",
                    "Gate Release",
                    "",
                    "Duck On",
                    "Duck Thr",
                    "Duck Release",
                    "",
                ),
            },
        ),
        (
            "Noise/Wobble",
            {
                BANK_PARAMETERS_KEY: (
                    "Noise On",
                    "Noise Amt",
                    "Noise Mrph",
                    "",
                    "Wobble On",
                    "Wobble Amt",
                    "Wobble Mrph",
                    "",
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["LoungeLizard"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    "M Stiffness",
                    "M Force",
                    "Noise Amount",
                    "F Tine Vol",
                    "F Tone Vol",
                    "F Release",
                    "P Symmetry",
                    "Volume",
                ),
            },
        ),
        (
            "Mallet",
            {
                BANK_PARAMETERS_KEY: (
                    "M Stiffness",
                    "M Force",
                    "Noise Pitch",
                    "Noise Decay",
                    "Noise Amount",
                    "M Stiff < Vel",
                    "M Force < Vel",
                    "Volume",
                ),
            },
        ),
        (
            "Fork",
            {
                BANK_PARAMETERS_KEY: (
                    "F Tine Color",
                    "F Tine Decay",
                    "F Tine Vol",
                    "F Tone Vol",
                    "F Tone Decay",
                    "F Release",
                    "F Tine < Key",
                    "Volume",
                ),
            },
        ),
        ("Damper", {BANK_PARAMETERS_KEY: ("Damp Tone", "Damp Balance", "Damp Amount", "", "", "", "", "Volume")}),
        (
            "Pickup",
            {
                BANK_PARAMETERS_KEY: (
                    "P Symmetry",
                    "P Distance",
                    "P Amp In",
                    "P Amp Out",
                    "Pickup Model",
                    "P Amp < Key",
                    "",
                    "Volume",
                ),
            },
        ),
        (
            "Modulation",
            {
                BANK_PARAMETERS_KEY: (
                    "M Stiff < Vel",
                    "M Stiff < Key",
                    "M Force < Vel",
                    "M Force < Key",
                    "Noise < Key",
                    "F Tine < Key",
                    "P Amp < Key",
                    "Volume",
                ),
            },
        ),
        ("Global", {BANK_PARAMETERS_KEY: ("KB Stretch", "PB Range", "", "", "Voices", "Semitone", "Detune", "Volume")}),
    ),
)
BANK_DEFINITIONS["Erosion"] = IndexedDict(
    ((BANK_MAIN_KEY, {BANK_PARAMETERS_KEY: ("Mode", "Frequency", "Width", "Amount", "", "", "", "")}),),
)
BANK_DEFINITIONS["ProxyAudioEffectDevice"] = IndexedDict(
    ((BANK_MAIN_KEY, {BANK_PARAMETERS_KEY: ("Input Gain", "Output Gain", "Dry/Wet", "", "", "", "", "")}),),
)
BANK_DEFINITIONS["FilterDelay"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    "2 Filter Freq",
                    "2 Filter Width",
                    use("2 Time Delay").if_parameter("2 Delay Mode").has_value("Off").else_use("2 Beat Delay"),
                    "2 Feedback",
                    "1 Volume",
                    "2 Volume",
                    "3 Volume",
                    "Dry",
                ),
            },
        ),
        (
            "L Filter",
            {
                BANK_PARAMETERS_KEY: (
                    "1 Input On",
                    "1 Filter Freq",
                    "1 Filter Width",
                    "1 Feedback",
                    "1 Delay Mode",
                    use("1 Time Delay").if_parameter("1 Delay Mode").has_value("Off").else_use("1 Beat Delay"),
                    "1 Beat Swing",
                    "1 Volume",
                ),
            },
        ),
        (
            "L+R Filter",
            {
                BANK_PARAMETERS_KEY: (
                    "2 Input On",
                    "2 Filter Freq",
                    "2 Filter Width",
                    "2 Feedback",
                    "2 Delay Mode",
                    use("2 Time Delay").if_parameter("2 Delay Mode").has_value("Off").else_use("2 Beat Delay"),
                    "2 Beat Swing",
                    "2 Volume",
                ),
            },
        ),
        (
            "R Filter",
            {
                BANK_PARAMETERS_KEY: (
                    "3 Input On",
                    "3 Filter Freq",
                    "3 Filter Width",
                    "3 Feedback",
                    "3 Delay Mode",
                    use("3 Time Delay").if_parameter("3 Delay Mode").has_value("Off").else_use("3 Beat Delay"),
                    "3 Beat Swing",
                    "3 Volume",
                ),
            },
        ),
        ("Mix", {BANK_PARAMETERS_KEY: ("1 Pan", "2 Pan", "3 Pan", "", "1 Volume", "2 Volume", "3 Volume", "Dry")}),
    ),
)
BANK_DEFINITIONS["Flanger"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    "LFO Amount",
                    "Sync",
                    use("Frequency").if_parameter("Sync").has_value("Free").else_use("Sync Rate"),
                    "Delay Time",
                    "Hi Pass",
                    "Env. Modulation",
                    "Feedback",
                    "Dry/Wet",
                ),
            },
        ),
        (
            "Envelope",
            {
                BANK_PARAMETERS_KEY: (
                    "Env. Attack",
                    "Env. Release",
                    "Env. Modulation",
                    "Hi Pass",
                    "Delay Time",
                    "Feedback",
                    "Polarity",
                    "Dry/Wet",
                ),
            },
        ),
        (
            "LFO / S&H",
            {
                BANK_PARAMETERS_KEY: (
                    "LFO Amount",
                    "LFO Waveform",
                    "Sync",
                    use("Frequency").if_parameter("Sync").has_value("Free").else_use("Sync Rate"),
                    use("LFO Stereo Mode").if_parameter("Sync").has_value("Free").else_use("LFO Offset"),
                    use("LFO Width (Random)")
                    .if_parameter("LFO Waveform")
                    .has_value("S&H Width")
                    .else_use("LFO Phase")
                    .if_parameter("Sync")
                    .has_value("Sync")
                    .or_parameter("LFO Stereo Mode")
                    .has_value("Phase")
                    .else_use("LFO Spin"),
                    "",
                    "",
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["FrequencyShifter"] = IndexedDict(
    (
        (
            "FreqDrive",
            {
                BANK_PARAMETERS_KEY: (
                    "Mode",
                    use("Ring Mod Frequency").if_parameter("Mode").has_value("Ring Modulation").else_use("Coarse"),
                    "Wide",
                    "Fine",
                    "Drive On/Off",
                    "Drive",
                    "LFO Amount",
                    "Dry/Wet",
                ),
            },
        ),
        (
            "LFO / S&H",
            {
                BANK_PARAMETERS_KEY: (
                    "LFO Amount",
                    "LFO Waveform",
                    "Sync",
                    use("LFO Frequency").if_parameter("Sync").has_value("Free").else_use("Sync Rate"),
                    use("LFO Stereo Mode").if_parameter("Sync").has_value("Free").else_use("LFO Offset"),
                    use("LFO Width (Random)")
                    .if_parameter("LFO Waveform")
                    .has_value("S&H Width")
                    .else_use("LFO Phase")
                    .if_parameter("Sync")
                    .has_value("Sync")
                    .or_parameter("LFO Stereo Mode")
                    .has_value("Phase")
                    .else_use("LFO Spin"),
                    "",
                    "",
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["Gate"] = IndexedDict(
    (
        (
            "Gate",
            {
                BANK_PARAMETERS_KEY: (
                    "Threshold",
                    "Return",
                    "FlipMode",
                    "LookAhead",
                    "Attack",
                    "Hold",
                    "Release",
                    "Floor",
                ),
            },
        ),
        (
            "Sidechain",
            {
                BANK_PARAMETERS_KEY: (
                    "S/C On",
                    "S/C Gain",
                    "S/C Mix",
                    "S/C Listen",
                    "S/C EQ On",
                    "S/C EQ Type",
                    "S/C EQ Freq",
                    use("S/C EQ Gain")
                    .if_parameter("S/C EQ Type")
                    .has_value_in(("Low Shelf", "High Shelf", "Bell"))
                    .else_use("S/C EQ Q"),
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["GlueCompressor"] = IndexedDict(
    (
        (
            "Compression",
            {
                BANK_PARAMETERS_KEY: (
                    "Threshold",
                    "Ratio",
                    "Attack",
                    "Release",
                    "Peak Clip In",
                    "Range",
                    "Makeup",
                    "Dry/Wet",
                ),
            },
        ),
        (
            "Sidechain",
            {
                BANK_PARAMETERS_KEY: (
                    "S/C On",
                    "S/C Gain",
                    "S/C Mix",
                    "",
                    "S/C EQ On",
                    "S/C EQ Type",
                    "S/C EQ Freq",
                    use("S/C EQ Gain")
                    .if_parameter("S/C EQ Type")
                    .has_value_in(("Low Shelf", "High Shelf", "Bell"))
                    .else_use("S/C EQ Q"),
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["GrainDelay"] = IndexedDict(
    (
        (
            "Pitch",
            {
                BANK_PARAMETERS_KEY: (
                    "Frequency",
                    "Pitch",
                    "Delay Mode",
                    use("Time Delay").if_parameter("Delay Mode").has_value("Off").else_use("Beat Delay"),
                    "Random",
                    "Spray",
                    "Feedback",
                    "DryWet",
                ),
            },
        ),
        (
            "Time",
            {
                BANK_PARAMETERS_KEY: (
                    "Delay Mode",
                    use("Time Delay").if_parameter("Delay Mode").has_value("Off").else_use("Beat Delay"),
                    "Beat Swing",
                    "Feedback",
                    "",
                    "",
                    "",
                    "DryWet",
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["Hybrid"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    use("Send Gain").with_name("Send"),
                    "Routing",
                    "Blend",
                    use("Algo Type").with_name("Type"),
                    "Decay",
                    "Size",
                    use("")
                    .if_parameter("Algo Type")
                    .has_value("Prism")
                    .else_use("Ti Waveform")
                    .with_name("Wave")
                    .if_parameter("Algo Type")
                    .has_value("Tides")
                    .else_use("Modulation")
                    .with_name("Mod"),
                    "Dry/Wet",
                ),
            },
        ),
        (
            "Algorithm Pg1",
            {
                BANK_PARAMETERS_KEY: (
                    use("Algo Type").with_name("Type"),
                    "Decay",
                    "Size",
                    use("DH Shape")
                    .with_name("Shape")
                    .if_parameter("Algo Type")
                    .has_value("Dark Hall")
                    .else_use("Qz Low Damp")
                    .with_name("Low Damping")
                    .if_parameter("Algo Type")
                    .has_value("Quartz")
                    .else_use("Sh Pitch Shift")
                    .with_name("Pitch")
                    .if_parameter("Algo Type")
                    .has_value("Shimmer")
                    .else_use("Ti Tide")
                    .with_name("Tide")
                    .if_parameter("Algo Type")
                    .has_value("Tides")
                    .else_use("Pr High Mult")
                    .with_name("High Mult"),
                    use("DH BassMult")
                    .with_name("Bass Mult")
                    .if_parameter("Algo Type")
                    .has_value("Dark Hall")
                    .else_use("Qz Distance")
                    .with_name("Distance")
                    .if_parameter("Algo Type")
                    .has_value("Quartz")
                    .else_use("Sh Shimmer")
                    .with_name("Shimmer")
                    .if_parameter("Algo Type")
                    .has_value("Shimmer")
                    .else_use("Ti Rate")
                    .with_name("Rate")
                    .if_parameter("Algo Type")
                    .has_value("Tides")
                    .else_use("Pr X Over")
                    .with_name("X Over"),
                    use("")
                    .if_parameter("Algo Type")
                    .has_value("Prism")
                    .else_use("Ti Waveform")
                    .with_name("Wave")
                    .if_parameter("Algo Type")
                    .has_value("Tides")
                    .else_use("Modulation")
                    .with_name("Mod"),
                    use("Width").with_name("Stereo"),
                    "Dry/Wet",
                ),
            },
        ),
        (
            "Algorithm Pg2",
            {
                BANK_PARAMETERS_KEY: (
                    use("Algo Type").with_name("Type"),
                    use("Pr Low Mult")
                    .with_name("Low Mult")
                    .if_parameter("Algo Type")
                    .has_value("Prism")
                    .else_use("Damping"),
                    use("DH Bass X")
                    .with_name("Bass X")
                    .if_parameter("Algo Type")
                    .has_value("Dark Hall")
                    .else_use("Diffusion")
                    .if_parameter("Algo Type")
                    .has_value_in(("Quartz", "Shimmer"))
                    .else_use("Ti Phase")
                    .with_name("Phase")
                    .if_parameter("Algo Type")
                    .has_value("Tides")
                    .else_use(""),
                    use("EQ Pre Algo").with_name("Pre Algo"),
                    use("Send Gain").with_name("Send"),
                    "Blend",
                    "Routing",
                    "Dry/Wet",
                ),
            },
        ),
        (
            "EQ Pg1",
            {
                BANK_PARAMETERS_KEY: (
                    use("EQ Pre Algo").with_name("Pre Algo"),
                    use("EQ Low Type").with_name("Low Type"),
                    use("EQ Low Freq").with_name("Low Freq"),
                    use("EQ Low Gain")
                    .with_name("Low Gain")
                    .if_parameter("EQ Low Type")
                    .has_value("Shelf")
                    .else_use("EQ Low Slope")
                    .with_name("Low Slope"),
                    use("EQ High Type").with_name("High Type"),
                    use("EQ High Freq").with_name("High Freq"),
                    use("EQ High Gain")
                    .with_name("High Gain")
                    .if_parameter("EQ High Type")
                    .has_value("Shelf")
                    .else_use("EQ High Slope")
                    .with_name("High Slope"),
                    "Dry/Wet",
                ),
            },
        ),
        (
            "EQ Pg2",
            {
                BANK_PARAMETERS_KEY: (
                    use("EQ Pre Algo").with_name("Pre Algo"),
                    use("EQ P1 Freq").with_name("2 Freq"),
                    use("EQ P1 Q").with_name("2 Q"),
                    use("EQ P1 Gain").with_name("2 Gain"),
                    use("EQ P2 Freq").with_name("3 Freq"),
                    use("EQ P2 Q").with_name("3 Q"),
                    use("EQ P2 Gain").with_name("3 Gain"),
                    "Dry/Wet",
                ),
            },
        ),
        (
            "Global",
            {
                BANK_PARAMETERS_KEY: (
                    use("Send Gain").with_name("Send"),
                    use("P.Dly Sync").with_name("Predelay Type"),
                    use("P.Dly Time")
                    .with_name("Predelay")
                    .if_parameter("P.Dly Sync")
                    .has_value("Off")
                    .else_use("P.Dly 16th")
                    .with_name("Predelay"),
                    use("P.Dly Fb Time")
                    .with_name("Predelay FB")
                    .if_parameter("P.Dly Sync")
                    .has_value("Off")
                    .else_use("P.Dly Fb 16th")
                    .with_name("Predelay FB"),
                    "Vintage",
                    use("Width").with_name("Stereo"),
                    "Blend",
                    "Dry/Wet",
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["InstrumentImpulse"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    "Global Time",
                    "Global Transpose",
                    "1 Transpose",
                    "2 Transpose",
                    "3 Transpose",
                    "4 Transpose",
                    "5 Transpose",
                    "6 Transpose",
                ),
            },
        ),
        (
            "Pad 1",
            {
                BANK_PARAMETERS_KEY: (
                    "1 Start",
                    "1 Transpose",
                    "1 Stretch Factor",
                    "1 Saturator Drive",
                    "1 Filter Freq",
                    "1 Filter Res",
                    "1 Pan",
                    "1 Volume",
                ),
            },
        ),
        (
            "Pad 2",
            {
                BANK_PARAMETERS_KEY: (
                    "2 Start",
                    "2 Transpose",
                    "2 Stretch Factor",
                    "2 Saturator Drive",
                    "2 Filter Freq",
                    "2 Filter Res",
                    "2 Pan",
                    "2 Volume",
                ),
            },
        ),
        (
            "Pad 3",
            {
                BANK_PARAMETERS_KEY: (
                    "3 Start",
                    "3 Transpose",
                    "3 Stretch Factor",
                    "3 Saturator Drive",
                    "3 Filter Freq",
                    "3 Filter Res",
                    "3 Pan",
                    "3 Volume",
                ),
            },
        ),
        (
            "Pad 4",
            {
                BANK_PARAMETERS_KEY: (
                    "4 Start",
                    "4 Transpose",
                    "4 Stretch Factor",
                    "4 Saturator Drive",
                    "4 Filter Freq",
                    "4 Filter Res",
                    "4 Pan",
                    "4 Volume",
                ),
            },
        ),
        (
            "Pad 5",
            {
                BANK_PARAMETERS_KEY: (
                    "5 Start",
                    "5 Transpose",
                    "5 Stretch Factor",
                    "5 Saturator Drive",
                    "5 Filter Freq",
                    "5 Filter Res",
                    "5 Pan",
                    "5 Volume",
                ),
            },
        ),
        (
            "Pad 6",
            {
                BANK_PARAMETERS_KEY: (
                    "6 Start",
                    "6 Transpose",
                    "6 Stretch Factor",
                    "6 Saturator Drive",
                    "6 Filter Freq",
                    "6 Filter Res",
                    "6 Pan",
                    "6 Volume",
                ),
            },
        ),
        (
            "Pad 7",
            {
                BANK_PARAMETERS_KEY: (
                    "7 Start",
                    "7 Transpose",
                    "7 Stretch Factor",
                    "7 Saturator Drive",
                    "7 Filter Freq",
                    "7 Filter Res",
                    "7 Pan",
                    "7 Volume",
                ),
            },
        ),
        (
            "Pad 8",
            {
                BANK_PARAMETERS_KEY: (
                    "8 Start",
                    "8 Transpose",
                    "8 Stretch Factor",
                    "8 Saturator Drive",
                    "8 Filter Freq",
                    "8 Filter Res",
                    "8 Pan",
                    "8 Volume",
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["Limiter"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {BANK_PARAMETERS_KEY: ("Gain", "Ceiling", "Link Channels", "Lookahead", "Auto", "Release time", "", "")},
        ),
    ),
)
BANK_DEFINITIONS["Looper"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    "State",
                    "Speed",
                    "Reverse",
                    "Quantization",
                    "Monitor",
                    "Song Control",
                    "Tempo Control",
                    "Feedback",
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["MultibandDynamics"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    use("Above Threshold (Low)").if_parameter("Band Activator (Low)").has_value("On"),
                    use("Above Ratio (Low)").if_parameter("Band Activator (Low)").has_value("On"),
                    use("Above Threshold (Mid)").if_parameter("Band Activator (Mid)").has_value("On"),
                    use("Above Ratio (Mid)").if_parameter("Band Activator (Mid)").has_value("On"),
                    use("Above Threshold (High)").if_parameter("Band Activator (High)").has_value("On"),
                    use("Above Ratio (High)").if_parameter("Band Activator (High)").has_value("On"),
                    "Master Output",
                    "Amount",
                ),
            },
        ),
        (
            "Low Band",
            {
                BANK_PARAMETERS_KEY: (
                    "Band Activator (Low)",
                    "Input Gain (Low)",
                    use("Below Threshold (Low)").if_parameter("Band Activator (Low)").has_value("On"),
                    use("Below Ratio (Low)").if_parameter("Band Activator (Low)").has_value("On"),
                    use("Above Threshold (Low)").if_parameter("Band Activator (Low)").has_value("On"),
                    use("Above Ratio (Low)").if_parameter("Band Activator (Low)").has_value("On"),
                    "Attack Time (Low)",
                    "Release Time (Low)",
                ),
            },
        ),
        (
            "Mid Band",
            {
                BANK_PARAMETERS_KEY: (
                    "Band Activator (Mid)",
                    "Input Gain (Mid)",
                    use("Below Threshold (Mid)").if_parameter("Band Activator (Mid)").has_value("On"),
                    use("Below Ratio (Mid)").if_parameter("Band Activator (Mid)").has_value("On"),
                    use("Above Threshold (Mid)").if_parameter("Band Activator (Mid)").has_value("On"),
                    use("Above Ratio (Mid)").if_parameter("Band Activator (Mid)").has_value("On"),
                    "Attack Time (Mid)",
                    "Release Time (Mid)",
                ),
            },
        ),
        (
            "High Band",
            {
                BANK_PARAMETERS_KEY: (
                    "Band Activator (High)",
                    "Input Gain (High)",
                    use("Below Threshold (High)").if_parameter("Band Activator (High)").has_value("On"),
                    use("Below Ratio (High)").if_parameter("Band Activator (High)").has_value("On"),
                    use("Above Threshold (High)").if_parameter("Band Activator (High)").has_value("On"),
                    use("Above Ratio (High)").if_parameter("Band Activator (High)").has_value("On"),
                    "Attack Time (High)",
                    "Release Time (High)",
                ),
            },
        ),
        (
            "Mix & Split",
            {
                BANK_PARAMETERS_KEY: (
                    "Output Gain (Low)",
                    "Low-Mid Crossover",
                    "Output Gain (Mid)",
                    "Mid-High Crossover",
                    "Output Gain (High)",
                    "Peak/RMS Mode",
                    "Amount",
                    "Master Output",
                ),
            },
        ),
        (
            "Sidechain",
            {BANK_PARAMETERS_KEY: ("S/C On", "S/C Mix", "S/C Gain", "", "Time Scaling", "Soft Knee On/Off", "", "")},
        ),
    ),
)
BANK_DEFINITIONS["MidiNoteLength"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    use("Trigger Source").with_name("Trigger Src").if_parameter("Latch").has_value("Off"),
                    use("Sync On").if_parameter("Latch").has_value("Off"),
                    use("Synced Length")
                    .if_parameter("Sync On")
                    .has_value("On")
                    .and_parameter("Latch")
                    .has_value("Off")
                    .else_use("Time Length")
                    .if_parameter("Latch")
                    .has_value("Off"),
                    use("Gate").if_parameter("Latch").has_value("Off"),
                    use("Release Velocity").with_name("Rel Vel").if_parameter("Latch").has_value("Off"),
                    use("Decay Time").if_parameter("Latch").has_value("Off"),
                    use("Decay Key Scale").if_parameter("Latch").has_value("Off"),
                    "Latch",
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["Operator"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    "Filter Freq",
                    "Filter Res",
                    use("A Coarse").if_parameter("A Fix On ").has_value("Off").else_use("A Fix Freq"),
                    use("A Fine").if_parameter("A Fix On ").has_value("Off").else_use("A Fix Freq Mul"),
                    use("B Coarse").if_parameter("B Fix On ").has_value("Off").else_use("B Fix Freq"),
                    use("B Fine").if_parameter("B Fix On ").has_value("Off").else_use("B Fix Freq Mul"),
                    "Osc-B Level",
                    "Volume",
                ),
            },
        ),
        (
            "Oscillator A",
            {
                BANK_PARAMETERS_KEY: (
                    "Ae Attack",
                    "Ae Decay",
                    "Ae Sustain",
                    "Ae Release",
                    use("A Coarse").if_parameter("A Fix On ").has_value("Off").else_use("A Fix Freq"),
                    use("A Fine").if_parameter("A Fix On ").has_value("Off").else_use("A Fix Freq Mul"),
                    "Osc-A Lev < Vel",
                    "Osc-A Level",
                ),
            },
        ),
        (
            "Oscillator B",
            {
                BANK_PARAMETERS_KEY: (
                    "Be Attack",
                    "Be Decay",
                    "Be Sustain",
                    "Be Release",
                    use("B Coarse").if_parameter("B Fix On ").has_value("Off").else_use("B Fix Freq"),
                    use("B Fine").if_parameter("B Fix On ").has_value("Off").else_use("B Fix Freq Mul"),
                    "Osc-B Lev < Vel",
                    "Osc-B Level",
                ),
            },
        ),
        (
            "Oscillator C",
            {
                BANK_PARAMETERS_KEY: (
                    "Ce Attack",
                    "Ce Decay",
                    "Ce Sustain",
                    "Ce Release",
                    use("C Coarse").if_parameter("C Fix On ").has_value("Off").else_use("C Fix Freq"),
                    use("C Fine").if_parameter("C Fix On ").has_value("Off").else_use("C Fix Freq Mul"),
                    "Osc-C Lev < Vel",
                    "Osc-C Level",
                ),
            },
        ),
        (
            "Oscillator D",
            {
                BANK_PARAMETERS_KEY: (
                    "De Attack",
                    "De Decay",
                    "De Sustain",
                    "De Release",
                    use("D Coarse").if_parameter("D Fix On ").has_value("Off").else_use("D Fix Freq"),
                    use("D Fine").if_parameter("D Fix On ").has_value("Off").else_use("D Fix Freq Mul"),
                    "Osc-D Lev < Vel",
                    "Osc-D Level",
                ),
            },
        ),
        (
            "LFO",
            {
                BANK_PARAMETERS_KEY: (
                    "Le Attack",
                    "Le Decay",
                    "Le Sustain",
                    "Le Release",
                    use("LFO Sync").if_parameter("LFO Range").has_value("Sync").else_use("LFO Rate"),
                    "LFO Amt",
                    "LFO Type",
                    "LFO R < K",
                ),
            },
        ),
        (
            "Filter",
            {
                BANK_PARAMETERS_KEY: (
                    "Fe Attack",
                    "Fe Decay",
                    "Fe Sustain",
                    "Fe Release",
                    "Filter Freq",
                    "Filter Res",
                    "Fe R < Vel",
                    "Fe Amount",
                ),
            },
        ),
        (
            "Pitch Modulation",
            {
                BANK_PARAMETERS_KEY: (
                    "Pe Attack",
                    "Pe Decay",
                    "Pe Sustain",
                    "Pe Release",
                    "Pe Init",
                    "Glide Time",
                    "Pe Amount",
                    "Spread",
                ),
            },
        ),
        (
            "Routing",
            {
                BANK_PARAMETERS_KEY: (
                    "Time < Key",
                    "Panorama",
                    "Pan < Key",
                    "Pan < Rnd",
                    "Algorithm",
                    "Time",
                    "Tone",
                    "Volume",
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["Overdrive"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    "Filter Freq",
                    "Filter Width",
                    "Drive",
                    "Tone",
                    "Preserve Dynamics",
                    "",
                    "",
                    "Dry/Wet",
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["Pedal"] = IndexedDict(
    (
        ("Global", {BANK_PARAMETERS_KEY: ("Type", "Gain", "Output", "Bass", "Mid", "Treble", "Sub", "Dry/Wet")}),
        ("EQ", {BANK_PARAMETERS_KEY: ("", "", "", "Bass", "Mid", "Treble", "", "Mid Freq")}),
    ),
)
BANK_DEFINITIONS["Phaser"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    "Poles",
                    "Frequency",
                    "Feedback",
                    "Env. Modulation",
                    "LFO Amount",
                    "LFO Sync",
                    use("LFO Frequency").if_parameter("LFO Sync").has_value("Free").else_use("LFO Sync Rate"),
                    "Dry/Wet",
                ),
            },
        ),
        (
            "Envelope",
            {
                BANK_PARAMETERS_KEY: (
                    "Poles",
                    "Type",
                    "Color",
                    "Frequency",
                    "Feedback",
                    "Env. Modulation",
                    "Env. Attack",
                    "Env. Release",
                ),
            },
        ),
        (
            "LFO",
            {
                BANK_PARAMETERS_KEY: (
                    "LFO Amount",
                    "LFO Waveform",
                    "LFO Sync",
                    use("LFO Frequency").if_parameter("LFO Sync").has_value("Free").else_use("LFO Sync Rate"),
                    use("LFO Stereo Mode").if_parameter("LFO Sync").has_value("Free").else_use("LFO Offset"),
                    use("LFO Width (Random)")
                    .if_parameter("LFO Waveform")
                    .has_value("S&H Width")
                    .else_use("LFO Phase")
                    .if_parameter("LFO Sync")
                    .has_value("Sync")
                    .or_parameter("LFO Stereo Mode")
                    .has_value("Phase")
                    .else_use("LFO Spin"),
                    "",
                    "",
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["PhaserNew"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    "Mode",
                    use("Mod Rate")
                    .with_name("Rate")
                    .if_parameter("Mod Sync")
                    .has_value("On")
                    .else_use("Mod Freq")
                    .with_name("Freq"),
                    "Amount",
                    "Feedback",
                    use("Notches")
                    .if_parameter("Mode")
                    .has_value("Phaser")
                    .else_use("Flange Time")
                    .with_name("Time")
                    .if_parameter("Mode")
                    .has_value("Flanger")
                    .else_use("Doubler Time")
                    .with_name("Time")
                    .if_parameter("Mode")
                    .has_value("Doubler"),
                    "Warmth",
                    use("Output Gain").with_name("Gain"),
                    "Dry/Wet",
                ),
            },
        ),
        (
            "Details",
            {
                BANK_PARAMETERS_KEY: (
                    "Mode",
                    use("Mod Rate")
                    .with_name("Rate")
                    .if_parameter("Mod Sync")
                    .has_value("On")
                    .else_use("Mod Freq")
                    .with_name("Freq"),
                    "Amount",
                    "Feedback",
                    use("Notches")
                    .if_parameter("Mode")
                    .has_value("Phaser")
                    .else_use("Flange Time")
                    .with_name("Time")
                    .if_parameter("Mode")
                    .has_value("Flanger")
                    .else_use("Doubler Time")
                    .with_name("Time")
                    .if_parameter("Mode")
                    .has_value("Doubler"),
                    use("FB Inv"),
                    use("Output Gain").with_name("Gain"),
                    "Dry/Wet",
                ),
            },
        ),
        (
            "LFO2",
            {
                BANK_PARAMETERS_KEY: (
                    use("Lfo Blend").with_name("LFO2 Mix"),
                    use("Mod Wave").with_name("Waveform"),
                    use("Duty Cycle").with_name("Duty"),
                    use("Spin Enabled").with_name("Spin On/Off"),
                    use("Spin").if_parameter("Spin Enabled").has_value("On").else_use("Mod Phase").with_name("Phase"),
                    use("Mod Sync 2").with_name("LFO2 Sync"),
                    use("Mod Rate 2")
                    .with_name("LFO2 Rate")
                    .if_parameter("Mod Sync 2")
                    .has_value("On")
                    .else_use("Mod Freq 2")
                    .with_name("LFO2 Freq"),
                    "Dry/Wet",
                ),
            },
        ),
        (
            "LFO2 Env",
            {
                BANK_PARAMETERS_KEY: (
                    use("Lfo Blend").with_name("LFO2 Mix"),
                    use("Env Enabled").with_name("Env Follow"),
                    "Env Amount",
                    use("Env Attack").with_name("Attack"),
                    use("Env Release").with_name("Release"),
                    use("Safe Freq").with_name("Safe Bass"),
                    use("Output Gain").with_name("Gain"),
                    "Dry/Wet",
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["MidiPitcher"] = IndexedDict(
    ((BANK_MAIN_KEY, {BANK_PARAMETERS_KEY: ("Pitch", "Range", "Lowest", "", "", "", "", "")}),),
)
BANK_DEFINITIONS["MidiRandom"] = IndexedDict(
    ((BANK_MAIN_KEY, {BANK_PARAMETERS_KEY: ("Chance", "Choices", "Mode", "Interval", "Sign", "", "", "")}),),
)
BANK_DEFINITIONS["Redux"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    "Bit On",
                    "Bit Depth",
                    "Sample Mode",
                    use("Sample Hard").if_parameter("Sample Mode").has_value("Hard").else_use("Sample Soft"),
                    "",
                    "",
                    "",
                    "",
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["Redux2"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    "Sample Rate",
                    "Jitter",
                    "Bit Depth",
                    use("Quantizer Shape").with_name("Shape"),
                    "DC Shift",
                    "Pre-Filter On",
                    "Post-Filter",
                    "Dry/Wet",
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["Resonator"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {BANK_PARAMETERS_KEY: ("Frequency", "Decay", "Color", "I Gain", "II Gain", "III Gain", "Width", "Dry/Wet")},
        ),
        ("Global", {BANK_PARAMETERS_KEY: ("Mode", "Decay", "Const", "Color", "", "Width", "Global Gain", "Dry/Wet")}),
        ("Filter", {BANK_PARAMETERS_KEY: ("Filter On", "Frequency", "Filter Type", "", "", "", "", "")}),
        (
            "Mode I & II",
            {BANK_PARAMETERS_KEY: ("I On", "I Note", "I Tune", "I Gain", "II On", "II Pitch", "II Tune", "II Gain")},
        ),
        (
            "Mode III & IV",
            {
                BANK_PARAMETERS_KEY: (
                    "III On",
                    "III Pitch",
                    "III Tune",
                    "III Gain",
                    "IV On",
                    "IV Pitch",
                    "IV Tune",
                    "IV Gain",
                ),
            },
        ),
        ("Mode V", {BANK_PARAMETERS_KEY: ("V On", "V Pitch", "V Tune", "V Gain", "", "", "", "")}),
        ("Mix", {BANK_PARAMETERS_KEY: ("I Gain", "II Gain", "III Gain", "IV Gain", "V Gain", "", "", "")}),
        ("Pitch", {BANK_PARAMETERS_KEY: ("I Note", "II Pitch", "III Pitch", "IV Pitch", "V Pitch", "", "", "")}),
    ),
)
BANK_DEFINITIONS["Reverb"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    "Predelay",
                    "In Filter Freq",
                    "Chorus Amount",
                    "Stereo Image",
                    "Room Size",
                    "Decay Time",
                    "Diffuse Level",
                    "Dry/Wet",
                ),
            },
        ),
        (
            "Global",
            {
                BANK_PARAMETERS_KEY: (
                    "Chorus On",
                    "Chorus Rate",
                    "Chorus Amount",
                    "Density",
                    "Freeze On",
                    "Flat On",
                    "Reflect Level",
                    "Diffuse Level",
                ),
            },
        ),
        (
            "Diffusion Network",
            {
                BANK_PARAMETERS_KEY: (
                    "HiFilter On",
                    "HiFilter Freq",
                    "HiShelf Gain",
                    "LowShelf On",
                    "LowShelf Freq",
                    "LowShelf Gain",
                    "Diffusion",
                    "Scale",
                ),
            },
        ),
        (
            "Input/Reflections",
            {
                BANK_PARAMETERS_KEY: (
                    "In LowCut On",
                    "In HighCut On",
                    "In Filter Freq",
                    "In Filter Width",
                    "ER Spin On",
                    "ER Spin Rate",
                    "ER Spin Amount",
                    "ER Shape",
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["MultiSampler"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    "Filter Freq",
                    "Filter Res",
                    "Fe < Env",
                    "Fe Decay",
                    "Ve Attack",
                    "Ve Release",
                    "Transpose",
                    "Volume",
                ),
            },
        ),
        (
            "Volume",
            {
                BANK_PARAMETERS_KEY: (
                    "Volume",
                    "Ve Attack",
                    "Ve Decay",
                    "Ve Sustain",
                    "Ve Release",
                    "Vol < Vel",
                    "Ve R < Vel",
                    "Time",
                ),
            },
        ),
        (
            "Filter",
            {
                BANK_PARAMETERS_KEY: (
                    "Filter Type",
                    use("Filter Morph").if_parameter("Filter Type").has_value("Morph").else_use("Filter Drive"),
                    "Filter Freq",
                    "Filter Res",
                    "Filt < Vel",
                    "Filt < Key",
                    "Fe < Env",
                    "Shaper Amt",
                ),
            },
        ),
        (
            "Filter Envelope",
            {
                BANK_PARAMETERS_KEY: (
                    "Fe Attack",
                    "Fe Peak",
                    "Fe Decay",
                    "Fe Sustain",
                    "Fe Release",
                    "Fe End",
                    "Fe Mode",
                    use("Fe Loop").if_parameter("Fe Mode").has_value("Loop").else_use("Fe Retrig"),
                ),
            },
        ),
        (
            "LFO 1",
            {
                BANK_PARAMETERS_KEY: (
                    "L 1 On",
                    "L 1 Wave",
                    "L 1 Sync",
                    use("L 1 Sync Rate").if_parameter("L 1 Sync").has_value("Sync").else_use("L 1 Rate"),
                    "Vol < LFO",
                    "Filt < LFO",
                    "Pan < LFO",
                    "Pitch < LFO",
                ),
            },
        ),
        (
            "LFO 2",
            {
                BANK_PARAMETERS_KEY: (
                    "L 2 On",
                    "L 2 Wave",
                    "L 2 Sync",
                    use("L 2 Sync Rate").if_parameter("L 2 Sync").has_value("Sync").else_use("L 2 Rate"),
                    "L 2 R < Key",
                    "L 2 Offset",
                    "L 2 St Mode",
                    use("L 2 Spin").if_parameter("L 2 St Mode").has_value("Spin").else_use("L 2 Phase"),
                ),
            },
        ),
        (
            "LFO 3",
            {
                BANK_PARAMETERS_KEY: (
                    "L 3 On",
                    "L 3 Wave",
                    "L 3 Sync",
                    use("L 3 Sync Rate").if_parameter("L 3 Sync").has_value("Sync").else_use("L 3 Rate"),
                    "L 3 R < Key",
                    "L 3 Offset",
                    "L 3 St Mode",
                    use("L 3 Spin").if_parameter("L 3 St Mode").has_value("Spin").else_use("L 3 Phase"),
                ),
            },
        ),
        (
            "Oscillator",
            {
                BANK_PARAMETERS_KEY: (
                    "O Mode",
                    "O Volume",
                    "O Coarse",
                    "O Fine",
                    "Oe Attack",
                    "Oe Decay",
                    "Oe Sustain",
                    "Oe Release",
                ),
            },
        ),
        (
            "Pitch",
            {
                BANK_PARAMETERS_KEY: (
                    "Transpose",
                    "Spread",
                    "Pe < Env",
                    "Pe Attack",
                    "Pe Peak",
                    "Pe Decay",
                    "Pe Sustain",
                    "Pe Release",
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["Saturator"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {BANK_PARAMETERS_KEY: ("Drive", "Type", "Color", "Base", "Frequency", "Width", "Depth", "Output")},
        ),
        (
            "Waveshaper",
            {
                BANK_PARAMETERS_KEY: (
                    "Type",
                    use("WS Drive").if_parameter("Type").has_value("Waveshaper"),
                    use("WS Curve").if_parameter("Type").has_value("Waveshaper"),
                    use("WS Depth").if_parameter("Type").has_value("Waveshaper"),
                    use("WS Lin").if_parameter("Type").has_value("Waveshaper"),
                    use("WS Damp").if_parameter("Type").has_value("Waveshaper"),
                    use("WS Period").if_parameter("Type").has_value("Waveshaper"),
                    "Dry/Wet",
                ),
            },
        ),
        ("Output", {BANK_PARAMETERS_KEY: ("", "", "", "", "", "Soft Clip", "Output", "Dry/Wet")}),
    ),
)
BANK_DEFINITIONS["MidiScale"] = IndexedDict(
    ((BANK_MAIN_KEY, {BANK_PARAMETERS_KEY: ("Base", "Transpose", "Range", "Lowest", "Fold", "", "", "")}),),
)
BANK_DEFINITIONS["Shifter"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    "Mode",
                    use("Pitch Coarse")
                    .with_name("Coarse")
                    .if_parameter("Mode")
                    .has_value("Pitch")
                    .else_use("FShift Coarse")
                    .with_name("Coarse")
                    .if_parameter("Mode")
                    .has_value("Freq")
                    .else_use("RM Coarse")
                    .with_name("Coarse"),
                    use("Pitch Fine")
                    .with_name("Spread")
                    .if_parameter("Mode")
                    .has_value("Pitch")
                    .and_parameter("Wide")
                    .has_value("On")
                    .else_use("Pitch Fine")
                    .with_name("Fine")
                    .if_parameter("Mode")
                    .has_value("Pitch")
                    .else_use("Mod Fine")
                    .with_name("Spread")
                    .if_parameter("Wide")
                    .has_value("On")
                    .else_use("Mod Fine")
                    .with_name("Fine"),
                    use("Lfo Waveform").with_name("Waveform"),
                    "Lfo Sync",
                    use("Lfo S. Rate")
                    .with_name("S. Rate")
                    .if_parameter("Lfo Sync")
                    .has_value("On")
                    .else_use("Lfo Rate Hz")
                    .with_name("Rate"),
                    use("Lfo Amount St")
                    .with_name("Amount")
                    .if_parameter("Mode")
                    .has_value("Pitch")
                    .else_use("Lfo Amount Hz")
                    .with_name("Amount"),
                    "Dry/Wet",
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["OriginalSimpler"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    "Filter Freq",
                    "Filter Res",
                    "S Start",
                    "S Length",
                    "Ve Attack",
                    "Ve Release",
                    "Transpose",
                    "Volume",
                ),
            },
        ),
        (
            "Amplitude",
            {
                BANK_PARAMETERS_KEY: (
                    "Ve Attack",
                    "Ve Decay",
                    "Ve Sustain",
                    "Ve Release",
                    "S Start",
                    "S Loop Length",
                    "S Length",
                    "S Loop Fade",
                ),
            },
        ),
        (
            "Filter",
            {
                BANK_PARAMETERS_KEY: (
                    "Fe Attack",
                    "Fe Decay",
                    "Fe Sustain",
                    "Fe Release",
                    "Filter Freq",
                    "Filter Res",
                    "Filt < Vel",
                    "Fe < Env",
                ),
            },
        ),
        (
            "LFO",
            {
                BANK_PARAMETERS_KEY: (
                    "L Attack",
                    use("L Rate").if_parameter("L Sync").has_value("Free").else_use("L Sync Rate"),
                    "L R < Key",
                    "L Wave",
                    "Vol < LFO",
                    "Filt < LFO",
                    "Pitch < LFO",
                    "Pan < LFO",
                ),
            },
        ),
        (
            "Pitch Modifiers",
            {
                BANK_PARAMETERS_KEY: (
                    "Pe Attack",
                    "Pe Decay",
                    "Pe Sustain",
                    "Pe Release",
                    "Glide Time",
                    "Spread",
                    "Pan",
                    "Volume",
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["Transmute"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    use("Transpose")
                    .if_parameter("Pitch Mode")
                    .has_value("MIDI")
                    .else_use("Freq. Hz")
                    .with_name("Freq")
                    .if_parameter("Hz/Note Mode")
                    .has_value("ModulationHertz")
                    .else_use("Note")
                    .with_name("Freq"),
                    "Hz/Note Mode",
                    "Decay",
                    "Stretch",
                    "Shift",
                    "Unison",
                    use("Input Send Gain").with_name("Input Send"),
                    use("Dry Wet").with_name("Dry/Wet"),
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["Spectral"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    "Frozen",
                    use("Delay Time Seconds")
                    .with_name("Del. Time")
                    .if_parameter("Delay Dly. Unit")
                    .has_value("Time")
                    .else_use("Delay Time Divisions")
                    .with_name("Del. Time")
                    .if_parameter("Delay Dly. Unit")
                    .has_value("Notes")
                    .else_use("Delay Time Sixteenths")
                    .with_name("Del. Time")
                    .if_parameter("Delay Dly. Unit")
                    .has_value("16th")
                    .else_use("Delay Time Sixteenths")
                    .with_name("Del. Time")
                    .if_parameter("Delay Dly. Unit")
                    .has_value("16th Triplet")
                    .else_use("Delay Time Sixteenths")
                    .with_name("Del. Time"),
                    use("Delay Feedback").with_name("Feedback"),
                    use("Delay Frequency Shift").with_name("Shift"),
                    use("Delay Tilt").with_name("Tilt"),
                    use("Delay Spray").with_name("Spray"),
                    use("Delay Mask").with_name("Mask"),
                    use("Dry Wet").with_name("Dry/Wet"),
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["StringStudio"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    "Filter Freq",
                    "Filter Reso",
                    "Filter Type",
                    "Exciter Type",
                    "E Pos",
                    "String Decay",
                    "Str Damping",
                    "Volume",
                ),
            },
        ),
        (
            "Exciter and String",
            {
                BANK_PARAMETERS_KEY: (
                    "Exciter Type",
                    "String Decay",
                    "Str Inharmon",
                    "Str Damping",
                    "Exc ForceMassProt",
                    "Exc FricStiff",
                    "Exc Velocity",
                    "E Pos",
                ),
            },
        ),
        (
            "Damper",
            {
                BANK_PARAMETERS_KEY: (
                    "Damper On",
                    "Damper Mass",
                    "D Stiffness",
                    "D Velocity",
                    "Damp Pos",
                    "D Damping",
                    "D Pos < Vel",
                    "D Pos Abs",
                ),
            },
        ),
        (
            "Termination and Pickup",
            {
                BANK_PARAMETERS_KEY: (
                    "Term On/Off",
                    "Term Mass",
                    "Term Fng Stiff",
                    "Term Fret Stiff",
                    "Pickup On/Off",
                    "Pickup Pos",
                    "T Mass < Vel",
                    "T Mass < Key",
                ),
            },
        ),
        (
            "Body",
            {
                BANK_PARAMETERS_KEY: (
                    "Body On/Off",
                    "Body Type",
                    "Body Size",
                    "Body Decay",
                    "Body Low-Cut",
                    "Body High-Cut",
                    "Body Mix",
                    "Volume",
                ),
            },
        ),
        (
            "Vibrato",
            {
                BANK_PARAMETERS_KEY: (
                    "Vibrato On/Off",
                    "Vib Delay",
                    "Vib Fade-In",
                    "Vib Speed",
                    "Vib Amount",
                    "Vib < ModWh",
                    "Vib Error",
                    "Volume",
                ),
            },
        ),
        (
            "Filter",
            {
                BANK_PARAMETERS_KEY: (
                    "Filter On/Off",
                    "Filter Type",
                    "Filter Freq",
                    "Filter Reso",
                    "Freq < Env",
                    "Freq < LFO",
                    "Reso < Env",
                    "Reso < LFO",
                ),
            },
        ),
        (
            "Envelope and LFO",
            {
                BANK_PARAMETERS_KEY: (
                    "FEG On/Off",
                    "FEG Attack",
                    "FEG Decay",
                    "FEG Sustain",
                    "FEG Release",
                    "LFO On/Off",
                    "LFO Shape",
                    use("LFO SyncRate").if_parameter("LFO Sync On").has_value("Beat").else_use("LFO Speed"),
                ),
            },
        ),
        (
            "Global",
            {
                BANK_PARAMETERS_KEY: (
                    "Unison On/Off",
                    "Uni Detune",
                    "Porta On/Off",
                    "Porta Time",
                    "Voices",
                    "Octave",
                    "Semitone",
                    "Volume",
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["StereoGain"] = IndexedDict(
    (
        (
            "Utility",
            {
                BANK_PARAMETERS_KEY: (
                    "Left Inv",
                    "Right Inv",
                    "Channel Mode",
                    use("Stereo Width").if_parameter("Stereo Width").is_available(True).else_use("Mid/Side Balance"),
                    "Mono",
                    "Balance",
                    use("Gain").if_parameter("Gain").is_available(True).else_use("Gain (Legacy)"),
                    "Mute",
                ),
            },
        ),
        (
            "Low Freq",
            {
                BANK_PARAMETERS_KEY: (
                    use("Bass Mono").if_parameter("Bass Mono").is_available(True),
                    use("Bass Freq").if_parameter("Bass Freq").is_available(True),
                    "",
                    "DC Filter",
                    "",
                    "",
                    "",
                    "",
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["MidiVelocity"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {BANK_PARAMETERS_KEY: ("Mode", "Drive", "Compand", "Out Hi", "Out Low", "Range", "Lowest", "Random")},
        ),
    ),
)
BANK_DEFINITIONS["Vinyl"] = IndexedDict(
    (
        (
            "Global",
            {
                BANK_PARAMETERS_KEY: (
                    "Tracing On",
                    "Tracing Drive",
                    "Tracing Freq.",
                    "Tracing Width",
                    "Pinch On",
                    "Global Drive",
                    "Crackle Density",
                    "Crackle Volume",
                ),
            },
        ),
        (
            "Pinch",
            {
                BANK_PARAMETERS_KEY: (
                    "Pinch On",
                    "Pinch Soft On",
                    "Pinch Mono On",
                    "Pinch Width",
                    "Pinch Drive",
                    "Pinch Freq.",
                    "Crackle Density",
                    "Crackle Volume",
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["Vocoder"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    "Formant Shift",
                    "Attack Time",
                    "Release Time",
                    "Unvoiced Level",
                    "Gate Threshold",
                    "Filter Bandwidth",
                    "Envelope Depth",
                    "Dry/Wet",
                ),
            },
        ),
        (
            "Carrier",
            {
                BANK_PARAMETERS_KEY: (
                    "Noise Rate",
                    "Noise Crackle",
                    "Upper Pitch Detection",
                    "Lower Pitch Detection",
                    "Oscillator Pitch",
                    "Oscillator Waveform",
                    "",
                    "",
                ),
            },
        ),
        (
            "Global",
            {
                BANK_PARAMETERS_KEY: (
                    "Formant Shift",
                    "Attack Time",
                    "Release Time",
                    "Mono/Stereo",
                    "Output Level",
                    "Gate Threshold",
                    "Envelope Depth",
                    "Dry/Wet",
                ),
            },
        ),
        (
            "Filters/Voicing",
            {
                BANK_PARAMETERS_KEY: (
                    "Filter Bandwidth",
                    "Upper Filter Band",
                    "Lower Filter Band",
                    "Precise/Retro",
                    "Unvoiced Level",
                    "Unvoiced Sensitivity",
                    "Unvoiced Speed",
                    "Enhance",
                ),
            },
        ),
    ),
)
BANK_DEFINITIONS["InstrumentVector"] = IndexedDict(
    (
        (
            BANK_MAIN_KEY,
            {
                BANK_PARAMETERS_KEY: (
                    "Osc 1 Pos",
                    "Osc 1 Transp",
                    "Osc 2 Pos",
                    "Osc 2 Transp",
                    "Filter 1 Freq",
                    "Filter 1 Res",
                    "Global Mod Amount",
                    "Volume",
                ),
            },
        ),
        (
            "Oscillator 1",
            {
                BANK_PARAMETERS_KEY: (
                    "Osc 1 Transp",
                    "Osc 1 Detune",
                    "Osc 1 Pos",
                    "Osc 1 Effect 1",
                    "Osc 1 Effect 2",
                    "Osc 1 Pan",
                    "Osc 1 Gain",
                    "Osc 1 On",
                ),
            },
        ),
        (
            "Oscillator 2",
            {
                BANK_PARAMETERS_KEY: (
                    "Osc 2 Transp",
                    "Osc 2 Detune",
                    "Osc 2 Pos",
                    "Osc 2 Effect 1",
                    "Osc 2 Effect 2",
                    "Osc 2 Pan",
                    "Osc 2 Gain",
                    "Osc 2 On",
                ),
            },
        ),
        (
            "Filter 1",
            {
                BANK_PARAMETERS_KEY: (
                    "Filter 1 On",
                    "Filter 1 Freq",
                    "Filter 1 Res",
                    use("Filter 1 Morph").if_parameter("Filter 1 Type").has_value("Morph").else_use("Filter 1 Drive"),
                    "Filter 1 Type",
                    "Filter 1 Slope",
                    use("Filter 1 LP/HP")
                    .if_parameter("Filter 1 Type")
                    .has_value_in(("Lowpass", "Highpass"))
                    .else_use("Filter 1 BP/NO/Morph"),
                    "",
                ),
            },
        ),
        (
            "Filter 2",
            {
                BANK_PARAMETERS_KEY: (
                    "Filter 2 On",
                    "Filter 2 Freq",
                    "Filter 2 Res",
                    use("Filter 2 Morph").if_parameter("Filter 2 Type").has_value("Morph").else_use("Filter 2 Drive"),
                    "Filter 2 Type",
                    "Filter 2 Slope",
                    use("Filter 2 LP/HP")
                    .if_parameter("Filter 2 Type")
                    .has_value_in(("Lowpass", "Highpass"))
                    .else_use("Filter 2 BP/NO/Morph"),
                    "",
                ),
            },
        ),
        (
            "Amp Envelope",
            {
                BANK_PARAMETERS_KEY: (
                    "Amp Attack",
                    "Amp Decay",
                    "Amp Sustain",
                    "Amp Release",
                    "Amp A Slope",
                    "Amp D Slope",
                    "Amp R Slope",
                    "Amp Loop Mode",
                ),
            },
        ),
        (
            "Envelope 2/3",
            {
                BANK_PARAMETERS_KEY: (
                    "Env 2 Attack",
                    "Env 2 Decay",
                    "Env 2 Sustain",
                    "Env 2 Release",
                    "Env 3 Attack",
                    "Env 3 Decay",
                    "Env 3 Sustain",
                    "Env 3 Release",
                ),
            },
        ),
        (
            "LFO 1/2",
            {
                BANK_PARAMETERS_KEY: (
                    "LFO 1 Amount",
                    "LFO 1 Shape",
                    use("LFO 1 S. Rate").if_parameter("LFO 1 Sync").has_value("Tempo").else_use("LFO 1 Rate"),
                    "LFO 1 Sync",
                    "LFO 2 Amount",
                    "LFO 2 Shape",
                    use("LFO 2 S. Rate").if_parameter("LFO 2 Sync").has_value("Tempo").else_use("LFO 2 Rate"),
                    "LFO 2 Sync",
                ),
            },
        ),
        (
            "Global",
            {
                BANK_PARAMETERS_KEY: (
                    "Time",
                    "Global Mod Amount",
                    "Unison Amount",
                    "Transpose",
                    "Glide",
                    "Sub Gain",
                    "Sub Transpose",
                    "Volume",
                ),
            },
        ),
    ),
)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/_MxDCore/LomTypes.py
# Compiled at: 2023-12-21 15:35:35
# Size of source mod 2**32: 41914 bytes
import ast
import json
from collections import namedtuple

import Live
from _Framework import ControlElement, ControlSurface, ControlSurfaceComponent
from _Framework.Util import is_iterable
from ableton.v2.base import old_hasattr
from past.builtins import basestring

from _MxDCore.ControlSurfaceWrapper import ControlProxy, ControlSurfaceWrapper


class MFLPropertyFormats:
    Default, JSON = (0, 1)


_MFLProperty = namedtuple("MFLProperty", "name format to_json from_json min_epii_version")
SORT_KEYS = False


def MFLProperty(name, format=MFLPropertyFormats.Default, to_json=None, from_json=None, min_epii_version=(-1, -1)):
    return _MFLProperty(name, format, to_json, from_json, min_epii_version)


def data_dict_to_json(property_name, data_dict):
    return json.dumps({property_name: data_dict}, ensure_ascii=True, sort_keys=SORT_KEYS)


def json_to_data_dict(property_name, json_dict):
    data_dict = ast.literal_eval(json_dict)
    return data_dict.get(property_name, data_dict)


def warp_markers_to_json(obj):
    property_value = obj.warp_markers
    return data_dict_to_json("warp_markers", tuple([warp_marker_to_dict(t) for t in property_value]))


def warp_marker_to_dict(warp_marker):
    return {"sample_time": warp_marker.sample_time, "beat_time": warp_marker.beat_time}


def verify_routings_available_for_object(obj, prop_name):
    if isinstance(obj, Live.Track.Track):
        error_format = "'%s' not available on %s"
        song = obj.canonical_parent
        if obj == song.master_track:
            raise RuntimeError(error_format % (prop_name, "master track"))
        else:
            if "input" in prop_name:
                if obj.is_foldable:
                    raise RuntimeError(error_format % (prop_name, "group tracks"))
                else:
                    if obj in song.return_tracks:
                        raise RuntimeError(error_format % (prop_name, "return tracks"))


def routing_object_to_dict(routing_type):
    return {"display_name": routing_type.display_name, "identifier": hash(routing_type)}


def available_routing_objects_to_json(obj, property_name):
    verify_routings_available_for_object(obj, property_name)
    property_value = getattr(obj, property_name)
    return data_dict_to_json(property_name, tuple([routing_object_to_dict(t) for t in property_value]))


def available_routing_input_types_to_json(obj):
    return available_routing_objects_to_json(obj, "available_input_routing_types")


def available_routing_output_types_to_json(obj):
    return available_routing_objects_to_json(obj, "available_output_routing_types")


def available_routing_input_channels_to_json(obj):
    return available_routing_objects_to_json(obj, "available_input_routing_channels")


def available_routing_output_channels_to_json(obj):
    return available_routing_objects_to_json(obj, "available_output_routing_channels")


def available_routing_types_to_json(device):
    return available_routing_objects_to_json(device, "available_routing_types")


def available_routing_channels_to_json(device):
    return available_routing_objects_to_json(device, "available_routing_channels")


def routing_object_to_json(obj, property_name):
    verify_routings_available_for_object(obj, property_name)
    property_value = getattr(obj, property_name)
    return data_dict_to_json(property_name, routing_object_to_dict(property_value))


def routing_input_type_to_json(obj):
    return routing_object_to_json(obj, "input_routing_type")


def routing_output_type_to_json(obj):
    return routing_object_to_json(obj, "output_routing_type")


def routing_input_channel_to_json(obj):
    return routing_object_to_json(obj, "input_routing_channel")


def routing_output_channel_to_json(obj):
    return routing_object_to_json(obj, "output_routing_channel")


def routing_type_to_json(device):
    return routing_object_to_json(device, "routing_type")


def routing_channel_to_json(device):
    return routing_object_to_json(device, "routing_channel")


def json_to_routing_object(obj, property_name, json_dict):
    verify_routings_available_for_object(obj, property_name)
    objects = getattr(obj, f"available_{property_name}s", [])
    identifier = json_to_data_dict(property_name, json_dict)["identifier"]
    for routing_object in objects:
        if hash(routing_object) == identifier:
            return routing_object
    return None


def json_to_input_routing_type(obj, json_dict):
    return json_to_routing_object(obj, "input_routing_type", json_dict)


def json_to_output_routing_type(obj, json_dict):
    return json_to_routing_object(obj, "output_routing_type", json_dict)


def json_to_input_routing_channel(obj, json_dict):
    return json_to_routing_object(obj, "input_routing_channel", json_dict)


def json_to_output_routing_channel(obj, json_dict):
    return json_to_routing_object(obj, "output_routing_channel", json_dict)


def json_to_routing_type(device, json_dict):
    return json_to_routing_object(device, "routing_type", json_dict)


def json_to_routing_channel(device, json_dict):
    return json_to_routing_object(device, "routing_channel", json_dict)


_DEVICE_BASE_PROPS = [
    MFLProperty("can_have_chains"),
    MFLProperty("can_have_drum_pads"),
    MFLProperty("canonical_parent"),
    MFLProperty("class_display_name"),
    MFLProperty("class_name"),
    MFLProperty("is_active"),
    MFLProperty("latency_in_ms"),
    MFLProperty("latency_in_samples"),
    MFLProperty("name"),
    MFLProperty("parameters"),
    MFLProperty("store_chosen_bank"),
    MFLProperty("type"),
    MFLProperty("view"),
]
_DEVICE_VIEW_BASE_PROPS = [MFLProperty("canonical_parent"), MFLProperty("is_collapsed")]
_CHAIN_BASE_PROPS = [
    MFLProperty("canonical_parent"),
    MFLProperty("color"),
    MFLProperty("color_index"),
    MFLProperty("delete_device"),
    MFLProperty("devices"),
    MFLProperty("has_audio_input"),
    MFLProperty("has_audio_output"),
    MFLProperty("has_midi_input"),
    MFLProperty("has_midi_output"),
    MFLProperty("is_auto_colored"),
    MFLProperty("mixer_device"),
    MFLProperty("mute"),
    MFLProperty("muted_via_solo"),
    MFLProperty("name"),
    MFLProperty("solo"),
]
EXPOSED_TYPE_PROPERTIES = {
    Live.Application.Application: (
        MFLProperty("control_surfaces"),
        MFLProperty("current_dialog_button_count"),
        MFLProperty("current_dialog_message"),
        MFLProperty("get_bugfix_version"),
        MFLProperty("get_document"),
        MFLProperty("get_version_string"),
        MFLProperty("get_major_version"),
        MFLProperty("get_minor_version"),
        MFLProperty("open_dialog_count"),
        MFLProperty("press_current_dialog_button"),
        MFLProperty("average_process_usage"),
        MFLProperty("peak_process_usage"),
        MFLProperty("view"),
    ),
    Live.Application.Application.View: (
        MFLProperty("available_main_views"),
        MFLProperty("browse_mode"),
        MFLProperty("canonical_parent"),
        MFLProperty("focus_view"),
        MFLProperty("focused_document_view"),
        MFLProperty("hide_view"),
        MFLProperty("is_view_visible"),
        MFLProperty("scroll_view"),
        MFLProperty("show_view"),
        MFLProperty("toggle_browse"),
        MFLProperty("zoom_view"),
    ),
    Live.Chain.Chain: tuple(_CHAIN_BASE_PROPS),
    Live.ChainMixerDevice.ChainMixerDevice: (
        MFLProperty("canonical_parent"),
        MFLProperty("chain_activator"),
        MFLProperty("panning"),
        MFLProperty("sends"),
        MFLProperty("volume"),
    ),
    Live.Clip.Clip: (
        MFLProperty("add_new_notes", min_epii_version=(4, 3)),
        MFLProperty("add_warp_marker", min_epii_version=(4, 3)),
        MFLProperty("apply_note_modifications", min_epii_version=(4, 3)),
        MFLProperty("available_warp_modes"),
        MFLProperty("canonical_parent"),
        MFLProperty("clear_all_envelopes"),
        MFLProperty("clear_envelope"),
        MFLProperty("color"),
        MFLProperty("color_index"),
        MFLProperty("crop"),
        MFLProperty("deselect_all_notes"),
        MFLProperty("duplicate_loop"),
        MFLProperty("duplicate_notes_by_id"),
        MFLProperty("duplicate_region"),
        MFLProperty("end_marker"),
        MFLProperty("end_time"),
        MFLProperty("file_path"),
        MFLProperty("fire"),
        MFLProperty("gain"),
        MFLProperty("gain_display_string"),
        MFLProperty("get_all_notes_extended"),
        MFLProperty("get_notes"),
        MFLProperty("get_notes_by_id", min_epii_version=(4, 3)),
        MFLProperty("get_notes_extended", min_epii_version=(4, 3)),
        MFLProperty("get_selected_notes"),
        MFLProperty("get_selected_notes_extended", min_epii_version=(4, 3)),
        MFLProperty("groove"),
        MFLProperty("has_envelopes"),
        MFLProperty("has_groove"),
        MFLProperty("is_arrangement_clip"),
        MFLProperty("is_audio_clip"),
        MFLProperty("is_midi_clip"),
        MFLProperty("is_overdubbing"),
        MFLProperty("is_playing"),
        MFLProperty("is_recording"),
        MFLProperty("is_triggered"),
        MFLProperty("launch_mode"),
        MFLProperty("launch_quantization"),
        MFLProperty("legato"),
        MFLProperty("length"),
        MFLProperty("loop_end"),
        MFLProperty("loop_start"),
        MFLProperty("looping"),
        MFLProperty("move_playing_pos"),
        MFLProperty("move_warp_marker"),
        MFLProperty("muted"),
        MFLProperty("name"),
        MFLProperty("pitch_coarse"),
        MFLProperty("pitch_fine"),
        MFLProperty("playing_position"),
        MFLProperty("position"),
        MFLProperty("quantize"),
        MFLProperty("quantize_pitch"),
        MFLProperty("ram_mode"),
        MFLProperty("remove_notes"),
        MFLProperty("remove_notes_by_id", min_epii_version=(4, 3)),
        MFLProperty("remove_notes_extended", min_epii_version=(4, 3)),
        MFLProperty("remove_warp_marker"),
        MFLProperty("replace_selected_notes"),
        MFLProperty("sample_length"),
        MFLProperty("sample_rate"),
        MFLProperty("scrub"),
        MFLProperty("select_all_notes"),
        MFLProperty("select_notes_by_id"),
        MFLProperty("set_fire_button_state"),
        MFLProperty("set_notes"),
        MFLProperty("signature_denominator"),
        MFLProperty("signature_numerator"),
        MFLProperty("start_marker"),
        MFLProperty("start_time"),
        MFLProperty("stop"),
        MFLProperty("stop_scrub"),
        MFLProperty("velocity_amount"),
        MFLProperty("view"),
        MFLProperty(
            "warp_markers",
            format=(MFLPropertyFormats.JSON),
            to_json=warp_markers_to_json,
            min_epii_version=(4, 3),
        ),
        MFLProperty("warp_mode"),
        MFLProperty("warping"),
        MFLProperty("will_record_on_start"),
    ),
    Live.Clip.Clip.View: (
        MFLProperty("canonical_parent"),
        MFLProperty("grid_is_triplet"),
        MFLProperty("grid_quantization"),
        MFLProperty("hide_envelope"),
        MFLProperty("select_envelope_parameter"),
        MFLProperty("show_envelope"),
        MFLProperty("show_loop"),
    ),
    Live.Groove.Groove: (
        MFLProperty("base"),
        MFLProperty("canonical_parent"),
        MFLProperty("name"),
        MFLProperty("quantization_amount"),
        MFLProperty("random_amount"),
        MFLProperty("timing_amount"),
        MFLProperty("velocity_amount"),
    ),
    Live.ClipSlot.ClipSlot: (
        MFLProperty("canonical_parent"),
        MFLProperty("clip"),
        MFLProperty("color"),
        MFLProperty("color_index"),
        MFLProperty("controls_other_clips"),
        MFLProperty("create_clip"),
        MFLProperty("delete_clip"),
        MFLProperty("duplicate_clip_to"),
        MFLProperty("fire"),
        MFLProperty("has_clip"),
        MFLProperty("has_stop_button"),
        MFLProperty("is_group_slot"),
        MFLProperty("is_playing"),
        MFLProperty("is_recording"),
        MFLProperty("is_triggered"),
        MFLProperty("playing_status"),
        MFLProperty("set_fire_button_state"),
        MFLProperty("stop"),
        MFLProperty("will_record_on_start"),
    ),
    Live.CompressorDevice.CompressorDevice: (
        *_DEVICE_BASE_PROPS,
        MFLProperty(
            "available_input_routing_channels",
            format=MFLPropertyFormats.JSON,
            to_json=available_routing_input_channels_to_json,
            min_epii_version=(4, 3),
        ),
        MFLProperty(
            "available_input_routing_types",
            format=MFLPropertyFormats.JSON,
            to_json=available_routing_input_types_to_json,
            min_epii_version=(4, 3),
        ),
        MFLProperty(
            "input_routing_channel",
            format=MFLPropertyFormats.JSON,
            to_json=routing_input_channel_to_json,
            from_json=json_to_input_routing_channel,
            min_epii_version=(4, 3),
        ),
        MFLProperty(
            "input_routing_type",
            format=MFLPropertyFormats.JSON,
            to_json=routing_input_type_to_json,
            from_json=json_to_input_routing_type,
            min_epii_version=(4, 3),
        ),
    ),
    Live.Device.Device: tuple(_DEVICE_BASE_PROPS),
    Live.Device.Device.View: tuple(_DEVICE_VIEW_BASE_PROPS),
    Live.DeviceParameter.DeviceParameter: (
        MFLProperty("__str__"),
        MFLProperty("automation_state"),
        MFLProperty("canonical_parent"),
        MFLProperty("default_value"),
        MFLProperty("is_enabled"),
        MFLProperty("is_quantized"),
        MFLProperty("max"),
        MFLProperty("min"),
        MFLProperty("name"),
        MFLProperty("original_name"),
        MFLProperty("re_enable_automation"),
        MFLProperty("state"),
        MFLProperty("str_for_value"),
        MFLProperty("value"),
        MFLProperty("value_items"),
    ),
    Live.DeviceIO.DeviceIO: (
        MFLProperty(
            "available_routing_channels",
            format=(MFLPropertyFormats.JSON),
            to_json=available_routing_channels_to_json,
            min_epii_version=(4, 4),
        ),
        MFLProperty(
            "available_routing_types",
            format=(MFLPropertyFormats.JSON),
            to_json=available_routing_types_to_json,
            min_epii_version=(4, 4),
        ),
        MFLProperty("canonical_parent"),
        MFLProperty("default_external_routing_channel_is_none"),
        MFLProperty(
            "routing_channel",
            format=(MFLPropertyFormats.JSON),
            to_json=routing_channel_to_json,
            from_json=json_to_routing_channel,
            min_epii_version=(4, 4),
        ),
        MFLProperty(
            "routing_type",
            format=(MFLPropertyFormats.JSON),
            to_json=routing_type_to_json,
            from_json=json_to_routing_type,
            min_epii_version=(4, 4),
        ),
    ),
    Live.DrumChain.DrumChain: (*_CHAIN_BASE_PROPS, MFLProperty("out_note"), MFLProperty("choke_group")),
    Live.DrumPad.DrumPad: (
        MFLProperty("canonical_parent"),
        MFLProperty("chains"),
        MFLProperty("delete_all_chains"),
        MFLProperty("mute"),
        MFLProperty("name"),
        MFLProperty("note"),
        MFLProperty("solo"),
    ),
    Live.Eq8Device.Eq8Device: (
        *_DEVICE_BASE_PROPS,
        MFLProperty("edit_mode"),
        MFLProperty("global_mode"),
        MFLProperty("oversample"),
    ),
    Live.Eq8Device.Eq8Device.View: (*_DEVICE_VIEW_BASE_PROPS, MFLProperty("selected_band")),
    Live.DriftDevice.DriftDevice: (
        *_DEVICE_BASE_PROPS,
        MFLProperty("voice_mode_index"),
        MFLProperty("voice_mode_list"),
        MFLProperty("voice_count_index"),
        MFLProperty("voice_count_list"),
        MFLProperty("mod_matrix_shape_source_index"),
        MFLProperty("mod_matrix_shape_source_list"),
        MFLProperty("mod_matrix_lfo_source_index"),
        MFLProperty("mod_matrix_lfo_source_list"),
        MFLProperty("mod_matrix_filter_source_1_index"),
        MFLProperty("mod_matrix_filter_source_1_list"),
        MFLProperty("mod_matrix_filter_source_2_index"),
        MFLProperty("mod_matrix_filter_source_2_list"),
        MFLProperty("mod_matrix_pitch_source_1_index"),
        MFLProperty("mod_matrix_pitch_source_1_list"),
        MFLProperty("mod_matrix_pitch_source_2_index"),
        MFLProperty("mod_matrix_pitch_source_2_list"),
        MFLProperty("mod_matrix_source_1_index"),
        MFLProperty("mod_matrix_source_1_list"),
        MFLProperty("mod_matrix_source_2_index"),
        MFLProperty("mod_matrix_source_2_list"),
        MFLProperty("mod_matrix_source_3_index"),
        MFLProperty("mod_matrix_source_3_list"),
        MFLProperty("pitch_bend_range"),
    ),
    Live.HybridReverbDevice.HybridReverbDevice: (
        *_DEVICE_BASE_PROPS,
        MFLProperty("ir_attack_time"),
        MFLProperty("ir_category_index"),
        MFLProperty("ir_category_list"),
        MFLProperty("ir_decay_time"),
        MFLProperty("ir_file_index"),
        MFLProperty("ir_file_list"),
        MFLProperty("ir_size_factor"),
        MFLProperty("ir_time_shaping_on"),
    ),
    Live.MaxDevice.MaxDevice: (
        *_DEVICE_BASE_PROPS,
        MFLProperty("audio_inputs"),
        MFLProperty("audio_outputs"),
        MFLProperty("get_bank_count"),
        MFLProperty("get_bank_name"),
        MFLProperty("get_bank_parameters"),
        MFLProperty("midi_inputs"),
        MFLProperty("midi_outputs"),
    ),
    Live.MeldDevice.MeldDevice: (
        *_DEVICE_BASE_PROPS,
        MFLProperty("selected_engine"),
        MFLProperty("unison_voices"),
        MFLProperty("mono_poly"),
        MFLProperty("poly_voices"),
    ),
    Live.MixerDevice.MixerDevice: (
        MFLProperty("canonical_parent"),
        MFLProperty("crossfade_assign"),
        MFLProperty("crossfader"),
        MFLProperty("cue_volume"),
        MFLProperty("left_split_stereo"),
        MFLProperty("panning"),
        MFLProperty("panning_mode"),
        MFLProperty("right_split_stereo"),
        MFLProperty("sends"),
        MFLProperty("song_tempo"),
        MFLProperty("track_activator"),
        MFLProperty("volume"),
    ),
    Live.PluginDevice.PluginDevice: (*_DEVICE_BASE_PROPS, MFLProperty("presets"), MFLProperty("selected_preset_index")),
    Live.RackDevice.RackDevice: (
        *_DEVICE_BASE_PROPS,
        MFLProperty("add_macro"),
        MFLProperty("can_show_chains"),
        MFLProperty("chain_selector"),
        MFLProperty("chains"),
        MFLProperty("copy_pad"),
        MFLProperty("delete_selected_variation"),
        MFLProperty("drum_pads"),
        MFLProperty("has_drum_pads"),
        MFLProperty("has_macro_mappings"),
        MFLProperty("is_showing_chains"),
        MFLProperty("randomize_macros"),
        MFLProperty("recall_last_used_variation"),
        MFLProperty("recall_selected_variation"),
        MFLProperty("remove_macro"),
        MFLProperty("return_chains"),
        MFLProperty("selected_variation_index"),
        MFLProperty("store_variation"),
        MFLProperty("variation_count"),
        MFLProperty("visible_drum_pads"),
        MFLProperty("visible_macro_count"),
    ),
    Live.RackDevice.RackDevice.View: (
        *_DEVICE_VIEW_BASE_PROPS,
        MFLProperty("drum_pads_scroll_position"),
        MFLProperty("is_showing_chain_devices"),
        MFLProperty("selected_chain"),
        MFLProperty("selected_drum_pad"),
    ),
    Live.RoarDevice.RoarDevice: (
        *_DEVICE_BASE_PROPS,
        MFLProperty("routing_mode_index"),
        MFLProperty("routing_mode_list"),
        MFLProperty("env_listen"),
    ),
    Live.Sample.Sample: (
        MFLProperty("beats_granulation_resolution"),
        MFLProperty("beats_transient_envelope"),
        MFLProperty("beats_transient_loop_mode"),
        MFLProperty("canonical_parent"),
        MFLProperty("clear_slices"),
        MFLProperty("complex_pro_envelope"),
        MFLProperty("complex_pro_formants"),
        MFLProperty("end_marker"),
        MFLProperty("file_path"),
        MFLProperty("gain"),
        MFLProperty("gain_display_string"),
        MFLProperty("insert_slice"),
        MFLProperty("length"),
        MFLProperty("move_slice"),
        MFLProperty("remove_slice"),
        MFLProperty("reset_slices"),
        MFLProperty("sample_rate"),
        MFLProperty("slices"),
        MFLProperty("slicing_beat_division"),
        MFLProperty("slicing_region_count"),
        MFLProperty("slicing_sensitivity"),
        MFLProperty("slicing_style"),
        MFLProperty("start_marker"),
        MFLProperty("texture_flux"),
        MFLProperty("texture_grain_size"),
        MFLProperty("tones_grain_size"),
        MFLProperty("warp_mode"),
        MFLProperty(
            "warp_markers",
            format=(MFLPropertyFormats.JSON),
            to_json=warp_markers_to_json,
            min_epii_version=(4, 3),
        ),
        MFLProperty("warping"),
    ),
    Live.Scene.Scene: (
        MFLProperty("canonical_parent"),
        MFLProperty("clip_slots"),
        MFLProperty("color"),
        MFLProperty("color_index"),
        MFLProperty("fire"),
        MFLProperty("fire_as_selected"),
        MFLProperty("is_empty"),
        MFLProperty("is_triggered"),
        MFLProperty("name"),
        MFLProperty("set_fire_button_state"),
        MFLProperty("tempo"),
        MFLProperty("tempo_enabled"),
        MFLProperty("time_signature_numerator"),
        MFLProperty("time_signature_denominator"),
        MFLProperty("time_signature_enabled"),
    ),
    Live.ShifterDevice.ShifterDevice: (
        *_DEVICE_BASE_PROPS,
        MFLProperty("pitch_bend_range"),
        MFLProperty("pitch_mode_index"),
    ),
    Live.SimplerDevice.SimplerDevice: (
        *_DEVICE_BASE_PROPS,
        MFLProperty("can_warp_as"),
        MFLProperty("can_warp_double"),
        MFLProperty("can_warp_half"),
        MFLProperty("crop"),
        MFLProperty("guess_playback_length"),
        MFLProperty("multi_sample_mode"),
        MFLProperty("pad_slicing"),
        MFLProperty("playback_mode"),
        MFLProperty("playing_position"),
        MFLProperty("playing_position_enabled"),
        MFLProperty("retrigger"),
        MFLProperty("reverse"),
        MFLProperty("sample"),
        MFLProperty("slicing_playback_mode"),
        MFLProperty("voices"),
        MFLProperty("warp_as"),
        MFLProperty("warp_double"),
        MFLProperty("warp_half"),
    ),
    Live.SimplerDevice.SimplerDevice.View: (*_DEVICE_VIEW_BASE_PROPS, MFLProperty("selected_slice")),
    Live.Song.Song: (
        MFLProperty("appointed_device"),
        MFLProperty("arrangement_overdub"),
        MFLProperty("back_to_arranger"),
        MFLProperty("can_capture_midi"),
        MFLProperty("can_jump_to_next_cue"),
        MFLProperty("can_jump_to_prev_cue"),
        MFLProperty("can_redo"),
        MFLProperty("can_undo"),
        MFLProperty("capture_and_insert_scene"),
        MFLProperty("capture_midi"),
        MFLProperty("clip_trigger_quantization"),
        MFLProperty("continue_playing"),
        MFLProperty("count_in_duration"),
        MFLProperty("create_audio_track"),
        MFLProperty("create_midi_track"),
        MFLProperty("create_return_track"),
        MFLProperty("create_scene"),
        MFLProperty("cue_points"),
        MFLProperty("current_song_time"),
        MFLProperty("delete_return_track"),
        MFLProperty("delete_scene"),
        MFLProperty("delete_track"),
        MFLProperty("duplicate_scene"),
        MFLProperty("duplicate_track"),
        MFLProperty("exclusive_arm"),
        MFLProperty("exclusive_solo"),
        MFLProperty("find_device_position"),
        MFLProperty("force_link_beat_time"),
        MFLProperty("get_beats_loop_length"),
        MFLProperty("get_beats_loop_start"),
        MFLProperty("get_current_beats_song_time"),
        MFLProperty("get_current_smpte_song_time"),
        MFLProperty("groove_amount"),
        MFLProperty("groove_pool"),
        MFLProperty("is_ableton_link_enabled"),
        MFLProperty("is_ableton_link_start_stop_sync_enabled"),
        MFLProperty("is_counting_in"),
        MFLProperty("is_cue_point_selected"),
        MFLProperty("is_playing"),
        MFLProperty("jump_by"),
        MFLProperty("jump_to_next_cue"),
        MFLProperty("jump_to_prev_cue"),
        MFLProperty("last_event_time"),
        MFLProperty("loop"),
        MFLProperty("loop_length"),
        MFLProperty("loop_start"),
        MFLProperty("master_track"),
        MFLProperty("metronome"),
        MFLProperty("midi_recording_quantization"),
        MFLProperty("move_device"),
        MFLProperty("nudge_down"),
        MFLProperty("nudge_up"),
        MFLProperty("tempo_follower_enabled"),
        MFLProperty("overdub"),
        MFLProperty("play_selection"),
        MFLProperty("punch_in"),
        MFLProperty("punch_out"),
        MFLProperty("re_enable_automation"),
        MFLProperty("re_enable_automation_enabled"),
        MFLProperty("record_mode"),
        MFLProperty("redo"),
        MFLProperty("return_tracks"),
        MFLProperty("root_note"),
        MFLProperty("scale_intervals"),
        MFLProperty("scale_name"),
        MFLProperty("scenes"),
        MFLProperty("scrub_by"),
        MFLProperty("select_on_launch"),
        MFLProperty("session_automation_record"),
        MFLProperty("session_record"),
        MFLProperty("session_record_status"),
        MFLProperty("set_or_delete_cue"),
        MFLProperty("signature_denominator"),
        MFLProperty("signature_numerator"),
        MFLProperty("song_length"),
        MFLProperty("start_playing"),
        MFLProperty("start_time"),
        MFLProperty("stop_all_clips"),
        MFLProperty("stop_playing"),
        MFLProperty("swing_amount"),
        MFLProperty("tap_tempo"),
        MFLProperty("tempo"),
        MFLProperty("tracks"),
        MFLProperty("trigger_session_record"),
        MFLProperty("undo"),
        MFLProperty("view"),
        MFLProperty("visible_tracks"),
    ),
    Live.Song.Song.View: (
        MFLProperty("canonical_parent"),
        MFLProperty("detail_clip"),
        MFLProperty("draw_mode"),
        MFLProperty("follow_song"),
        MFLProperty("highlighted_clip_slot"),
        MFLProperty("select_device"),
        MFLProperty("selected_chain"),
        MFLProperty("selected_parameter"),
        MFLProperty("selected_scene"),
        MFLProperty("selected_track"),
    ),
    Live.GroovePool.GroovePool: (MFLProperty("canonical_parent"), MFLProperty("grooves")),
    Live.Song.CuePoint: (
        MFLProperty("canonical_parent"),
        MFLProperty("jump"),
        MFLProperty("name"),
        MFLProperty("time"),
    ),
    Live.Track.Track: (
        MFLProperty("arm"),
        MFLProperty("arrangement_clips"),
        MFLProperty(
            "available_input_routing_channels",
            format=(MFLPropertyFormats.JSON),
            to_json=available_routing_input_channels_to_json,
            min_epii_version=(4, 3),
        ),
        MFLProperty(
            "available_input_routing_types",
            format=(MFLPropertyFormats.JSON),
            to_json=available_routing_input_types_to_json,
            min_epii_version=(4, 3),
        ),
        MFLProperty(
            "available_output_routing_channels",
            format=(MFLPropertyFormats.JSON),
            to_json=available_routing_output_channels_to_json,
            min_epii_version=(4, 3),
        ),
        MFLProperty(
            "available_output_routing_types",
            format=(MFLPropertyFormats.JSON),
            to_json=available_routing_output_types_to_json,
            min_epii_version=(4, 3),
        ),
        MFLProperty("can_be_armed"),
        MFLProperty("can_be_frozen"),
        MFLProperty("can_show_chains"),
        MFLProperty("canonical_parent"),
        MFLProperty("clip_slots"),
        MFLProperty("color"),
        MFLProperty("color_index"),
        MFLProperty("current_input_routing"),
        MFLProperty("current_input_sub_routing"),
        MFLProperty("current_monitoring_state"),
        MFLProperty("current_output_routing"),
        MFLProperty("current_output_sub_routing"),
        MFLProperty("delete_clip"),
        MFLProperty("delete_device"),
        MFLProperty("devices"),
        MFLProperty("duplicate_clip_slot"),
        MFLProperty("duplicate_clip_to_arrangement"),
        MFLProperty("fired_slot_index"),
        MFLProperty("fold_state"),
        MFLProperty("group_track"),
        MFLProperty("has_audio_input"),
        MFLProperty("has_audio_output"),
        MFLProperty("has_midi_input"),
        MFLProperty("has_midi_output"),
        MFLProperty("implicit_arm"),
        MFLProperty("input_meter_left"),
        MFLProperty("input_meter_level"),
        MFLProperty("input_meter_right"),
        MFLProperty(
            "input_routing_channel",
            format=(MFLPropertyFormats.JSON),
            to_json=routing_input_channel_to_json,
            from_json=json_to_input_routing_channel,
            min_epii_version=(4, 3),
        ),
        MFLProperty(
            "input_routing_type",
            format=(MFLPropertyFormats.JSON),
            to_json=routing_input_type_to_json,
            from_json=json_to_input_routing_type,
            min_epii_version=(4, 3),
        ),
        MFLProperty("input_routings"),
        MFLProperty("input_sub_routings"),
        MFLProperty("is_foldable"),
        MFLProperty("is_frozen"),
        MFLProperty("is_grouped"),
        MFLProperty("is_part_of_selection"),
        MFLProperty("is_showing_chains"),
        MFLProperty("is_visible"),
        MFLProperty("jump_in_running_session_clip"),
        MFLProperty("mixer_device"),
        MFLProperty("mute"),
        MFLProperty("muted_via_solo"),
        MFLProperty("name"),
        MFLProperty("output_meter_left"),
        MFLProperty("output_meter_level"),
        MFLProperty("output_meter_right"),
        MFLProperty(
            "output_routing_channel",
            format=(MFLPropertyFormats.JSON),
            to_json=routing_output_channel_to_json,
            from_json=json_to_output_routing_channel,
            min_epii_version=(4, 3),
        ),
        MFLProperty(
            "output_routing_type",
            format=(MFLPropertyFormats.JSON),
            to_json=routing_output_type_to_json,
            from_json=json_to_output_routing_type,
            min_epii_version=(4, 3),
        ),
        MFLProperty("output_routings"),
        MFLProperty("output_sub_routings"),
        MFLProperty("performance_impact"),
        MFLProperty("playing_slot_index"),
        MFLProperty("solo"),
        MFLProperty("stop_all_clips"),
        MFLProperty("view"),
    ),
    Live.Track.Track.View: (
        MFLProperty("canonical_parent"),
        MFLProperty("device_insert_mode"),
        MFLProperty("is_collapsed"),
        MFLProperty("select_instrument"),
        MFLProperty("selected_device"),
    ),
    Live.SpectralResonatorDevice.SpectralResonatorDevice: (
        *_DEVICE_BASE_PROPS,
        MFLProperty("frequency_dial_mode"),
        MFLProperty("midi_gate"),
        MFLProperty("mod_mode"),
        MFLProperty("mono_poly"),
        MFLProperty("pitch_mode"),
        MFLProperty("pitch_bend_range"),
        MFLProperty("polyphony"),
    ),
    Live.WavetableDevice.WavetableDevice: (
        *_DEVICE_BASE_PROPS,
        MFLProperty("add_parameter_to_modulation_matrix"),
        MFLProperty("filter_routing"),
        MFLProperty("get_modulation_target_parameter_name"),
        MFLProperty("get_modulation_value"),
        MFLProperty("is_parameter_modulatable"),
        MFLProperty("mono_poly"),
        MFLProperty("oscillator_1_effect_mode"),
        MFLProperty("oscillator_1_wavetable_category"),
        MFLProperty("oscillator_1_wavetable_index"),
        MFLProperty("oscillator_1_wavetables"),
        MFLProperty("oscillator_2_effect_mode"),
        MFLProperty("oscillator_2_wavetable_category"),
        MFLProperty("oscillator_2_wavetable_index"),
        MFLProperty("oscillator_2_wavetables"),
        MFLProperty("oscillator_wavetable_categories"),
        MFLProperty("poly_voices"),
        MFLProperty("set_modulation_value"),
        MFLProperty("unison_mode"),
        MFLProperty("unison_voice_count"),
        MFLProperty("visible_modulation_target_names"),
    ),
}
PROPERTY_NAMES_EXCLUDED_FROM_DOCUMENTATION = ("canonical_parent",)
EXTRA_CS_FUNCTIONS = (
    "get_control",
    "get_control_names",
    "grab_control",
    "grab_midi",
    "release_control",
    "release_midi",
    "send_midi",
    "send_receive_sysex",
)
ENUM_TYPES = (
    Live.Song.Quantization,
    Live.Song.RecordingQuantization,
    Live.Song.CaptureMode,
    Live.Clip.GridQuantization,
    Live.DeviceParameter.AutomationState,
    Live.Groove.Base,
    Live.Sample.SlicingStyle,
    Live.Sample.SlicingBeatDivision,
)
TUPLE_TYPES = {
    "arrangement_clips": Live.Clip.Clip,
    "audio_inputs": Live.DeviceIO.DeviceIO,
    "audio_outputs": Live.DeviceIO.DeviceIO,
    "chains": Live.Chain.Chain,
    "clip_slots": Live.ClipSlot.ClipSlot,
    "components": ControlSurfaceComponent,
    "control_surfaces": ControlSurface,
    "controls": ControlElement,
    "cue_points": Live.Song.CuePoint,
    "devices": Live.Device.Device,
    "drum_pads": Live.DrumPad.DrumPad,
    "grooves": Live.Groove.Groove,
    "midi_inputs": Live.DeviceIO.DeviceIO,
    "midi_outputs": Live.DeviceIO.DeviceIO,
    "parameters": Live.DeviceParameter.DeviceParameter,
    "return_chains": Live.Chain.Chain,
    "return_tracks": Live.Track.Track,
    "scenes": Live.Scene.Scene,
    "sends": Live.DeviceParameter.DeviceParameter,
    "tracks": Live.Track.Track,
    "visible_drum_pads": Live.DrumPad.DrumPad,
    "visible_tracks": Live.Track.Track,
}
PROPERTY_TYPES = {
    "chain_activator": Live.DeviceParameter.DeviceParameter,
    "chain_selector": Live.DeviceParameter.DeviceParameter,
    "clip": Live.Clip.Clip,
    "crossfader": Live.DeviceParameter.DeviceParameter,
    "cue_volume": Live.DeviceParameter.DeviceParameter,
    "detail_clip": Live.Clip.Clip,
    "groove": Live.Groove.Groove,
    "groove_pool": Live.GroovePool.GroovePool,
    "group_track": Live.Track.Track,
    "highlighted_clip_slot": Live.ClipSlot.ClipSlot,
    "left_split_stereo": Live.DeviceParameter.DeviceParameter,
    "master_track": Live.Track.Track,
    "mixer_device": (Live.MixerDevice.MixerDevice, Live.ChainMixerDevice.ChainMixerDevice),
    "panning": Live.DeviceParameter.DeviceParameter,
    "right_split_stereo": Live.DeviceParameter.DeviceParameter,
    "sample": Live.Sample.Sample,
    "selected_chain": Live.Chain.Chain,
    "selected_device": Live.Device.Device,
    "selected_drum_pad": Live.DrumPad.DrumPad,
    "selected_parameter": Live.DeviceParameter.DeviceParameter,
    "selected_scene": Live.Scene.Scene,
    "selected_track": Live.Track.Track,
    "song_tempo": Live.DeviceParameter.DeviceParameter,
    "track_activator": Live.DeviceParameter.DeviceParameter,
    "volume": Live.DeviceParameter.DeviceParameter,
    "view": (
        Live.Application.Application.View,
        Live.Song.Song.View,
        Live.Track.Track.View,
        Live.Device.Device.View,
        Live.RackDevice.RackDevice.View,
        Live.Clip.Clip.View,
    ),
}
LIVE_APP = "live_app"
LIVE_SET = "live_set"
CONTROL_SURFACES = "control_surfaces"
THIS_DEVICE = "this_device"
ROOT_KEYS = (THIS_DEVICE, CONTROL_SURFACES, LIVE_APP, LIVE_SET)


class LomAttributeError(AttributeError):
    pass


class LomObjectError(AttributeError):
    pass


class LomNoteOperationWarning(Exception):
    pass


class LomNoteOperationError(AttributeError):
    pass


def get_exposed_lom_types():
    return list(EXPOSED_TYPE_PROPERTIES.keys())


def get_exposed_properties_for_type(lom_type, epii_version=None):
    if epii_version is None:
        epii_version = (float("inf"), float("inf"))
    return [prop for prop in EXPOSED_TYPE_PROPERTIES.get(lom_type, []) if epii_version >= prop.min_epii_version]


def get_exposed_property_names_for_type(lom_type, epii_version=None):
    return [prop.name for prop in get_exposed_properties_for_type(lom_type, epii_version)]


def is_property_exposed_for_type(property_name, lom_type, epii_version):
    return property_name in get_exposed_property_names_for_type(lom_type, epii_version)


def get_exposed_property_info(lom_type, property_name, epii_version):
    properties = get_exposed_properties_for_type(lom_type, epii_version)
    prop = [p for p in properties if p.name == property_name]
    if not prop:
        return None
    return prop[0]


def get_exposed_properties_to_document_for_type(lom_type):
    properties = set(get_exposed_property_names_for_type(lom_type))
    if issubclass(lom_type, Live.Device.Device):
        if lom_type != Live.Device.Device:
            properties -= {prop.name for prop in _DEVICE_BASE_PROPS}
    if issubclass(lom_type, Live.Chain.Chain):
        if lom_type != Live.Chain.Chain:
            properties -= {prop.name for prop in _CHAIN_BASE_PROPS}
    if issubclass(lom_type, Live.Device.Device.View):
        if lom_type != Live.Device.Device.View:
            properties -= {prop.name for prop in _DEVICE_VIEW_BASE_PROPS}
    properties -= set(PROPERTY_NAMES_EXCLUDED_FROM_DOCUMENTATION)
    return list(properties)


def is_class(class_object):
    return isinstance(class_object, type) or old_hasattr(class_object, "__bases__")


def get_control_surfaces():
    return tuple(filter(lambda c: c is not None, Live.Application.get_application().control_surfaces))


def get_root_prop(external_device, prop_key):
    root_properties = {
        LIVE_APP: Live.Application.get_application,
        LIVE_SET: lambda: Live.Application.get_application().get_document(),
        CONTROL_SURFACES: get_control_surfaces,
    }
    if prop_key == THIS_DEVICE:
        return external_device
    return root_properties[prop_key]()


def cs_base_classes():
    from _Framework import ControlElement, ControlSurfaceComponent
    from ableton.v2.control_surface import Component as ControlSurfaceComponent2
    from ableton.v2.control_surface import ControlElement as ControlElement2
    from ableton.v3.control_surface import Component as ControlSurfaceComponent3

    return (
        ControlProxy,
        ControlSurfaceWrapper,
        ControlSurfaceComponent,
        ControlElement,
        ControlSurfaceComponent2,
        ControlElement2,
        ControlSurfaceComponent3,
    )


def is_control_surface(lom_object):
    return isinstance(lom_object, ControlSurfaceWrapper)


def is_lom_object(lom_object, lom_classes):
    return isinstance(lom_object, (Live.Base.Vector, cs_base_classes(), (*tuple(lom_classes), type(None))))


def is_cplusplus_lom_object(lom_object):
    return isinstance(lom_object, Live.LomObject.LomObject)


def is_object_iterable(obj):
    return not isinstance(obj, basestring) and is_iterable(obj) and not isinstance(obj, cs_base_classes())


def verify_object_property(lom_object, property_name, epii_version):
    raise_error = False
    if isinstance(lom_object, cs_base_classes()):
        if not old_hasattr(lom_object, property_name):
            raise_error = True
    else:
        if not is_property_exposed_for_type(property_name, type(lom_object), epii_version):
            raise_error = True
    if raise_error:
        raise LomAttributeError(f"'{lom_object.__class__.__name__}' object has no attribute '{property_name}'")
